# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Tue May 21 11:00:48 2019

Constructing and solving structural optimization problems

Idea: all the data is stored in OptimizationProblem object.
As different optimization solvers are applied, the methods of
OptimizationProblem class will be used to write the problem data in
a form suitable for each solver.

Classes:
    Variable
    IntegerVariable(Variable)
    BinaryVariable(IntegerVariable)
    IndexVariable(IntegerVariable)
    DiscreteVariable(Variable)


@author: kmela
"""

import copy
import time
from itertools import product

import numpy as np
import scipy.optimize as sciop

from metku.optimization.constants import XLB, XUB
from metku.optimization.constraints import Constraint, LinearConstraint, NonLinearConstraint
from metku.optimization.objective import ObjectiveFunction
from metku.optimization.opt_functions import NumGrad, conlin_fun, mma_fun
from metku.optimization.variables import BinaryVariable, IndexVariable
from metku.optimization.variables import Variable, DiscreteVariable, IntegerVariable


# from metku.eurocodes.en1991.loadIDs import LoadIDs

class OptimizationProblem:
    """ Class for defining and operating with optimization problems """

    def __init__(self, name="", variables=None, constraints=None, objective=None, \
                 gradient=None, hess=None, structure=None, profiles=None):
        """ Constructor

            Parameters:
            -----------
                :param name: name of the problem (string)
                :param variables: list of variables (Variable)
                :param constraints: list of constraints (Constraint)
                :param objective: objective function
                :param gradient: gradient of the objective function
                :param hess: Hessian of the objective function
                :param structure: structure to be optimized (optional).
                    If structural analysis is required in optimization, 'structure'
                    must have the method 'calculate' that performs structural analysis
                :param profiles: list of available profiles (optional)

                Variables:
                ----------
                :ivar name: name of the problem
                :ivar vars: variables (list)
                :ivar cons: constraints (list)
                :ivar obj: objective function
                :ivar grad: gradient of the objective function
                :ivar hess: Hessian of the objective function
                :ivar structure: structure to be optimized
                :ivar profiles: list of available profiles (optional)
                :ivar fea_done: indicates if FEM analysis has been done (bool)
                :ivar X: stored design variable values
                :ivar x0: initial point
                :ivar num_iters: number of iterations (int)
                :ivar num_fem_analyses: number of FEM analyses (int)
                :ivar fvals: number of function evaluations (int)
                :ivar states:
                :ivar gvals:
        """
        self.con_tol = 1e-4
        self.name = name
        if variables:
            self.__vars = []
            self.vars = variables
        else:
            self.__vars = []
        if constraints:
            self.cons = constraints
        else:
            self.cons = []
        self.obj = objective
        # Gradient and Hessian of the objective function
        # should also include the possibility that some of the variables
        # can be fixed!
        self.grad = gradient
        self.hess = hess
        self.structure = structure
        self.profiles = profiles
        self.fea_done = False
        self.X = None
        self.x0 = None
        self.num_iters = 0
        self.num_fem_analyses = 0
        self.fvals = []
        self.states = []
        self.gvals = []
    
    def xlb(self):
        """ Returns a numpy array of variable lower bounds """
        return np.array([x.lb for x in self.vars])

    def xub(self):
        """ Returns a numpy array of variable upper bounds """
        return np.array([x.ub for x in self.vars])

    def add(self, this):
        """
        Adds given object to the problem
        :param this: object to be added
        
        :type: Variable, Constraint, ObjectiveFunction
        """
        # VARIABLES
        if isinstance(this, Variable):
            if this.id is None:
                this.id = len(self.vars)
            if this not in self.__vars:
                self.__vars.append(this)

        # CONSTRAINTS
        elif isinstance(this, Constraint):
            if this not in self.cons:
            #if this not in self.cons and \
            #        (isinstance(this, NonLinearConstraint) and this.con not in [con.con for con in self.cons]):
                self.cons.append(this)
                this.problem = self

        # OBJECTIVES
        elif isinstance(this, ObjectiveFunction):
            # TODO: Multiple objectives
            self.obj = this
            this.problem = self
        else:
            raise ValueError(f"{this} must be either Variable, Constraint or ObjectiveFunction")

    @property
    def locked_vars(self):
        """
        Returns locked variables

        :return: list of locked variables (Variable objects, not values!)
        """
        return [var for var in self.__vars if var.locked]

    @property
    def all_vars(self):
        """
        Returns all variables

        :return: list of all variables
        """
        return self.__vars

    @property
    def vars(self):
        """
        Returns all unlocked variables

        :return: list of unlocked variables
        """
        return [var for var in self.__vars if not var.locked]

    @vars.setter
    def vars(self, vals):
        """
            Sets variables, to 'vals' (list of Variable objects)            
        """
        self.clear_vars()
        for val in vals:
            self.add(val)

    @property
    def fixed_vals(self):
        vals = []
        for var in self.__vars:
            if var.locked:
                vals.append(var.value)
            else:
                vals.append(None)
        return vals

    #@property
    def feasible(self,x=None):
        """
        Returns problem's feasibility at x
        :return: True/False
        """
        
        res = True

        if x is None:
            x = self.X

           
        for con in self.cons:
            g = con(x)             
            if con.type == '<':
                if g > self.con_tol:
                    #print('<= type inequality constraint infeasible')
                    #print(con,g)
                    res = False
                    break
            elif con.type == '>':
                if g < -self.con_tol:
                    #print('>= type inequality constraint infeasible')
                    #print(con,g)
                    res = False
                    break
            else:
                if abs(g) > self.con_tol:
                    #print('== type inequality constraint infeasible')
                    #print(con,g)
                    res = False
                    break
        
        return res
        #return np.all(self.eval_cons(self.X) <= self.con_tol)
    
    

    def type_of_vars(self):
        '''
        :return: "Variable", "IntegerVariable", "BinaryVariable",
        IndexVariable or "DiscreteVariable"
        '''
        var = self.vars[0]
        if isinstance(var, BinaryVariable):
            return "BinaryVariable"
        elif isinstance(var, IndexVariable):
            return "IndexVariable"
        elif isinstance(var, DiscreteVariable):
            return "DiscreteVariable"
        elif isinstance(var, IntegerVariable):
            return "IntegerVariable"
        elif isinstance(var, Variable):
            return "Variable"
        else:
            raise Exception("Error, type of type_of_vars is none of above")

    def clear_vars(self):
        """
            Deletes all variables
        """
        self.__vars.clear()
        
    def del_var(self,var):
        """
            Deletes variable 'var'

        """
        self.__vars.remove(var)
        
    def var_values(self):
        """
        Returns all current variable values
        -------

        """
        return np.array([var.value for var in self.vars])
    
    def random_values(self):
        """
        Generates random variable values within the range of allowable values
        --------
        
        Returns
        -------
        None.

        """
        xrand = np.zeros(self.nvars())
        rng = np.random.default_rng()

        for i, var in enumerate(self.vars):
            if isinstance(var,IntegerVariable) or isinstance(var,IndexVariable):
                xrand[i] = rng.integers(var.lb,var.ub+1)
            elif isinstance(var, DiscreteVariable):
                possible_values = var.values
                random_value = (var.ub - var.lb) * rng.random() + var.lb
                xrand[i] = min(possible_values, key=lambda x:abs(x-random_value))
            else:
                xrand[i] = (var.ub-var.lb)*rng.random() + var.lb
        return xrand

    def print_vars(self):
        """ Prints the current variable values """
        
        print("Current variables:")
        for var in self.vars:
            print("{0:s} = {1:5.4f}".format(var.name,var.value))

    def non_linear_constraint(self, *args, **kwargs):
        """
        Creates new NonLinearConstraint and adds it to cons -list

        :param args:
        :param kwargs:
        :return: NonLinearConstraint
        """
        con = NonLinearConstraint(*args, **kwargs)
        con.problem = self

        self.cons.append(con)
        return con

    def linear_constraint(self, *args, **kwargs):
        """
        Creates new LinearConstraint and adds it to cons -list
        :param args:
        :param kwargs:
        :return: LinearConstraint
        """

        con = LinearConstraint(*args, **kwargs)
        con.problem = self
        self.cons.append(con)
        return con

    #@lru_cache(CACHE_BOUND)
    def linearize(self, x, only_potential=False):
        """
        Linearizes problem around x

        Ax < B

        :param x:
        :param only_potential: True, if only potentially active constraints are included
        in the linearization
        :return:
        """
        #start = time.time()
        x = np.asarray(x)

        self.substitute_variables(x)

        """ Evaluate objective function at x """
        fx = self.obj(x)
        """ Evaluate nonlinear constraint functions at x """
        #b = self.eval_nonlin_cons(x)
        
        con_vals = self.eval_cons(x)
                
        """ Get number of nonlinear constraints """
        if only_potential:
            act_cons = [con for con in self.cons if con.potential]
            b = np.array([val for val, con in zip(con_vals,self.cons) if con.potential])
        else:
            act_cons = self.cons
            b = con_vals
            
        m = len(act_cons)
        #m = self.nnonlincons()
        """ Number of variables """
        n = self.nvars()
        """ Jacobian
            Rows of A are the gradients (transposed) of constraints
        """
        A = np.zeros((m, n))
        df = np.zeros(n)
        
        if self.grad is not None:
            df = self.grad(x)
                
        for i, var in enumerate(self.vars):
            if isinstance(var, BinaryVariable):
                h = 1
            elif isinstance(var, IndexVariable):
                h = 1
                prev_val = var.value
                if var.idx + h > var.ub:
                    h = -1
                var.substitute(var.idx + h)
            else:
                x[i] = var.value
                """ Get current value of variable """
                prev_val = var.value
                """ Step length """
                    
                h = max(1e-6 * abs(prev_val), 1e-8)
                var.substitute(prev_val + h)
                
            """ Get variable values """
            xh = [var.value for var in self.vars]
            #print(x,xh)
            if not isinstance(var,BinaryVariable) and np.linalg.norm(x-xh) == 0:
                print("Linearize: error with variable substitution - too small step size.")

            """ Evaluate objective function at x + hi*ei """
            if self.grad is None:
                if not isinstance(var, BinaryVariable):                 
                    f_val = self.obj(xh)
                    #print(f_val,fx)         
                    df[i] = (f_val - fx) / h
                    
            """ Evaluate constraint functions at x + hi*ei """
            #a = np.array([con(xh) for con in act_cons])
            dg = np.zeros(m)
            for j, con in enumerate(act_cons):
                if isinstance(con,LinearConstraint):
                    dg[j] = con.a[i]
                else:
                    if not isinstance(var,BinaryVariable):
                        dg[j] = (con(xh)-b[j])/h
                #a = self.eval_nonlin_cons(xh)
                #A[:, i] = (a - b) / h
            A[:, i] = dg
                
            """ Substitute the original value x[i] to current variable """
            var.substitute(prev_val)            
            

        #B = A @ x.T - b
        B = A.dot(x) - b
        # B = -b

        # Add linear constraints' values
        """
        for con in self.cons:
            if isinstance(con, LinearConstraint):
                A = np.vstack((A, con.a))
                if isinstance(con.b, Variable):
                    B = np.hstack((B, con.b.value))
                else:
                    B = np.hstack((B, con.b))
        """
        #end = time.time()
        # print("Linearization took: ", end - start, " s")

        return A, B, df, fx
    
    def conlin(self,x):
        """ Generates convex-linear approximation of the problem
            around point 'x'
            
            Assumption: all variables are positive
            
            Note: Linear function remain linear
        """
        
        Pconlin = OptimizationProblem(name="conlin")
        
        for i, var in enumerate(self.vars):
            """ Create variables for the conlin problem """
            Pconlin.add(Variable(name=var.name,lb=var.lb,ub=var.ub,value=x[i]))
        
        # Approximate objective
        if self.grad is None:
            dfx = NumGrad(self.obj,x)
        else:
            dfx = self.grad(x)
            
        Pconlin.obj = conlin_fun(x,self.obj(x),dfx)
        
        # Approximate constraints
        for con in self.cons:
            if isinstance(con,LinearConstraint):
                # Linear constraints are simply added
                Pconlin.add(con)
            else:
                # For nonlinear constraints, CONLIN approximation is made
                conlin_con = copy.copy(con)
                gx = con.con(x)
                dgx = con.df(x)
                conlin_con.con = conlin_fun(x,gx,dgx)
                
                Pconlin.add(conlin_con)
        
        return Pconlin
    
    def mma(self,x,U,L,mu=0.8):
        """
        Generates MMA approximation of the problem at point 'x' using the asymptotes 'U' and 'L'


        Parameters
        ----------
        x : numpy array
            Approximation point.
        U : numpy array
            Upper asymptote.
        L : numpy array
            Lower asymptote.

        Returns
        -------
        Optimization problem with nonlinear function replaced by their MMA approximations.

        """
        
        Pmma = OptimizationProblem(name="MMA")
        
        for i, var in enumerate(self.vars):
            """ Create variables for the mma problem """
            Pmma.add(Variable(name=var.name,lb=max(var.lb,L[i]+mu*(x[i]-L[i])),ub=min(var.ub,U[i]-mu*(U[i]-x[i])),value=x[i]))
        
        # Approximate objective
        if self.grad is None:
            dfx = NumGrad(self.obj,x)
        else:
            dfx = self.grad(x)
            
        Pmma.obj = mma_fun(x,self.obj(x),dfx,L,U)
        
        # Approximate constraints
        for con in self.cons:
            if isinstance(con,LinearConstraint):
                # Linear constraints are simply added
                Pmma.add(con)
            else:
                # For nonlinear constraints, MMA approximation is made
                mma_con = copy.copy(con)
                gx = con.con(x)
                dgx = con.df(x)
                mma_con.con = mma_fun(x,gx,dgx,L,U)
                
                Pmma.add(mma_con)
        
        return Pmma
        
    def nnonlincons(self):
        """ Returns number of non linear constraints"""
        non_lin_cons = [con for con in self.cons if
                        isinstance(con, NonLinearConstraint)]
        return len(non_lin_cons)

    def eval_nonlin_cons(self, X):
        """
        Calculates non linear constraints and returns their values as a numpy array
        :param X:
        :return:
        """
        g = []
        non_lin_cons = [con for con in self.cons if
                        isinstance(con, NonLinearConstraint)]
        start = time.time()
        for con in non_lin_cons:
            g.append(con(X))

        return np.asarray(g)

    #def fea(self, load_id=LoadIDs.ULS):
    def fea(self, load_id=None):
        """
        Runs finite element analysis on structure
        """
        if load_id is None:
            load_id = self.structure.load_ids[0]
        
        self.structure.calculate(load_id)
        self.fea_done = True
        self.num_fem_analyses += 1

    def __call__(self, x=None, prec=2, ncons=5):
        """ Call method evaluates the objective function and all constraints
            at x and returns their values
        """
        if x is None:
            x = [var.value for var in self.vars]
        self.substitute_variables(x)
        fx = self.obj(x)
        print("** {0} **".format(self.name))
        print(f'Solution is feasible: {self.feasible()}')

        #vals = [var.value for var in self.vars]
        #print(f"Optimal values: {vals}")

        print(f'X: {[round(val, prec) for val in x]}')

        if len(self.cons):
            g = self.eval_cons(x)
            print(F"Max constraint: {self.cons[np.argmax(g)].name}: {max(g):.{prec}f}")
        if len(x) < 100:
            print("Variables:")
            print("----------")
            for var in self.vars:
                if isinstance(var, IndexVariable):
                    print(f"{var.name} = {var.values[var.value]}")
                else:
                    print(f"{var.name} = {var.value:.{prec}f}")

            print("----------\n")

        print(f"Objective function = {fx:.{prec}f}\n")

        if len(self.cons):
            ncons = min(len(self.cons), ncons)
            print(f"{ncons} Maximum Constraints:")
            print("----------")

            idx = np.argpartition(g, -ncons)[-ncons:]
            for i in idx:
                gi = g[i]
                con = self.cons[i]
                print(f"{con.name}: {gi:.{prec}f} {con.type} 0")
            # for con in np.asarray(self.cons)[idx]:
            #     gi = con(x)
            #     print(f"{con.name}: {gi:.{prec}f} {con.type} 0")


    def eval_cons(self, x):
        """ Constraint evaluation

        """
        g = []

        for con in self.cons:
            g.append(con(x))

        return np.asarray(g)

    def eval_eq_con(self, x):
        """ Evaluates equality constraints """

        geq = []

        for con in self.cons:
            if con.type == "=":
                geq.append(con(x))

    def eval_ineq_con(self, x):
        """ Evaluate inequality constriants
            All constraints are evaluated in the form g(x) >= 0
        """

        g = []

        for con in self.cons:
            if con.type == ">":
                g.append(con(x))
            elif con.type == "<":
                g.append(-con(x))

    def nvars(self):
        """ Return number of variables """
        return len(self.vars)

    def nlincons(self):
        """ Number of nonlinear inequality constraints """

        return len(self.vars)

    def nlineqcons(self):
        """ Number of nonlinear equality constraints """

        return len(self.vars)

    def add_variable(self, var):
        """ Adds new variable to the problem """
        self.vars.append(var)

    def add_variables(self, variables):
        """ Adds new variables to the problem """
        for var in variables:
            self.vars.append(var)

    def add_constraint(self, con):
        """ Adds new constraint to the problem """
        self.cons.append(con)

    def add_constraints(self, constraints):
        """ Adds new constraints to the problem """
        for con in constraints:
            self.cons.append(con)

    def add_structure(self, structure):
        """ Add structure to the problem """
        self.structure = structure

    def var_bounds(self):
        """ Returns arrays lb and ub for variable bounds """
        lb = []
        ub = []

        for v in self.vars:
            lb.append(v.lb)
            ub.append(v.ub)

        return lb, ub

    # THIS METHOD IS NOT NEEDED!! USE substitute_variables INSTEAD!!
    def substitute_variable(self, i, xval):
        """ Substitutes the value 'xval' for variable i """
        var = self.vars[i]
        var.value = xval

        if var.target is not None:
            for obj in var.target['objects']:
                if var.target['property'] == 'A':
                    obj.A = xval
                elif var.target['property'] == 'IY':
                    obj.I[0] = xval
                elif var.target['property'] == 'IZ':
                    obj.I[1] = xval
                elif var.target['property'] == 'WELY':
                    obj.Wel[0] = xval
                elif var.target['property'] == 'WELZ':
                    obj.Wel[1] = xval
                elif var.target['property'] == 'WPLY':
                    obj.Wpl[0] = xval
                elif var.target['property'] == 'WPLZ':
                    obj.Wpl[1] = xval
                elif var.target['property'] == 'H':
                    obj.h = xval
                elif var.target['property'] == 'TW':
                    obj.tw = xval
                elif var.target['property'] == 'BF':
                    obj.b = xval
                elif var.target['property'] == 'TF':
                    obj.tf = xval
                elif var.target['property'] == 'PROFILE':
                    obj.profile = self.profiles(xval)

    def substitute_variables(self, xvals):
        """ Substitute variable values from xval to structure """

        """ Make sure all variable values are within lower and upper bounds """
        xvals = np.array([max(var.lb, min(x, var.ub)) for x, var in zip(xvals, self.vars)])
        #print(xvals)
        #print(self.var_values())
        #if np.linalg.norm(self.var_values()-xvals) > 1e-9:
        # Save starting point
        if not np.any(self.X):
            self.x0 = xvals.copy()

        # if np.any(self.X != xvals):
        self.X = xvals.copy()
        self.fea_done = False
        for x, var in zip(xvals, self.vars):
            var.substitute(x)
        #else:
            #print("No variable substitution.")
            #if self.nvars() < 10:
            #    print(self.var_values(),xvals)
            #
            # for i in range(len(xvals)):
            #     self.substitute_variable(i, xvals[i])

    # OBSOLETE! USE SOLVERS FROM solvers MODULE!
    def solve(self, solver="slsqp", **kwargs):
        """ Solve the optimization problem
            input:
                solver .. name of the solver
                **kwargs .. list of keyword argument
        """
        lb, ub = self.var_bounds()

        if solver == "slsqp":
            jac = "2-point"
            bnds = sciop.Bounds(lb, ub)

            if self.grad is not None:
                jac = self.grad

            if self.hess is not None:
                hess = self.hess

            """ Crerate constraints
            Constraints for COBYLA, SLSQP are defined as a 
            list of dictionaries. Each dictionary with fields:

                type: str

                Constraint type: ‘eq’ for equality, ‘ineq’ for inequality.

                fun: callable The function defining the constraint.

                jac: callable, optional. The Jacobian of fun (only for SLSQP).

                args: sequence, optional

                    Extra arguments to be passed to the function and Jacobian.

            """
            # cons = [{"type":"ineq", "fun":self.eval_ineq_con}]

            cons = []
            for con in self.cons:
                if con.type == "=":
                    new_con = {"type": "eq", "fun": con}
                elif con.type == "<":
                    new_con = {"type": "ineq", "fun": con.neg_call}
                else:
                    new_con = {"type": "ineq", "fun": con}

                cons.append(new_con)

            # print(cons)

            res = sciop.minimize(self.obj, kwargs["x0"], method=solver, \
                                 jac=jac, bounds=bnds, constraints=cons)
        elif solver == "trust-constr":
            jac = "2-point"
            hess = "2-point"
            bnds = sciop.Bounds(lb, ub)

            if self.grad is not None:
                jac = self.grad

            if self.hess is not None:
                hess = self.hess

            res = sciop.minimize(self.obj, kwargs["x0"], method=solver, \
                                 jac=jac, hess=hess, bounds=bnds)
        else:
            print("Solver unidentified.")
            res = None

        return res

    def discrete_neighborhood(self, x, k, method='nearest'):
        """ Creates a discrete neighborhood around point 'x' 
            input:
                k -- parameter used by the 'method'
                method .. procedure for defining the neigborhood
        """
        neighborhood = []
        
        for i, var in enumerate(self.vars):
            if method == 'nearest':
                if isinstance(var,DiscreteVariable):
                    values = np.array(var.values)
                    """ Sort values in ascending order of distance from 'x' 
                        np.argsort returns the indices of 'values'
                    """
                    sorted_ndx = np.argsort(abs(values-x[i]))
                    """ Get k actual values closest to x and transform
                        the Numpy array to list.
                        
                        Update the list of values for this variable
                    """
                    neighborhood.append(values[sorted_ndx[:k]].tolist())
                elif isinstance(var,IntegerVariable):
                    """ For integer variables, make a list of
                        integers between lb and ub
                        
                        The sorting is the same as for discrete variables,
                        but here only lb and ub need to be updated.
                    """
                    values = np.arange(var.lb,var.ub+1)
                    sorted_ndx = np.argsort(abs(values-x[i]))
                    new_values = values[sorted_ndx[:k]].tolist()
                    neighborhood.append(new_values)
                    #var.lb = min(new_values)
                    #var.ub = max(new_values)
                
        
        """
            Create iterable nList that contains the neighborhood points
            as a list of tuples
        """
        nList = product(*neighborhood)
        return nList
    
    def exact_penalty_fun(self,R=1.0):
        """ Exact penalty function 
            NOTE: Returns only the sum of constraint violations!
        """
        
        def penalty_fun(x):
            
            F = 0.0
            
            for con in self.cons:
                val = con(x)
                if con.type == '=':
                    """ For equality constraints, the penalty term
                        is simply the value of the constraint squared
                    """
                    F += abs(val)
                elif con.type == '<':
                    """ For less than or equal constraints, the penalty
                        term is max(0,val)
                    """
                    F += R*max(0,val)
                elif con.type == '>':
                    """ For geq type constraints, the penalty term is
                         min(0,val)
                    """
                    F += R*min(0,val)
            
            """ Variable bounds are treated separately """
            for (i,var) in enumerate(self.vars):                
                if var.lb > XLB:                    
                    #F += R*max(0,-1-x[i]/abs(var.lb))**2
                    F += R*max(0,-x[i]+var.lb)
                
                if var.ub < XUB:
                    #F += R*max(0,x[i]/abs(var.ub)-1)**2
                    F += R*max(0,x[i]-var.ub)
            
            return F
        
        return penalty_fun
            

    def exterior_penalty_fun(self,R=10,p=2):
        """ Return exterior quadratic penalty function """
        
        def penalty_fun(x):
            
            F = self.obj(x)
            
            for con in self.cons:
                val = con(x)
                if con.type == '=':
                    """ For equality constraints, the penalty term
                        is simply the value of the constraint squared
                    """
                    F += val**p
                elif con.type == '<':
                    """ For less than or equal constraints, the penalty
                        term is square of max(0,val)
                    """
                    F += R*max(0,val)**p
                elif con.type == '>':
                    """ For geq type constraints, the penalty term is
                        square of min(0,val)
                    """
                    F += R*min(0,val)**2
            
            
            for (i,var) in enumerate(self.vars):                
                if var.lb > XLB:                    
                    #F += R*max(0,-1-x[i]/abs(var.lb))**2
                    F += R*max(0,-x[i]+var.lb)**2
                
                if var.ub < XUB:
                    #F += R*max(0,x[i]/abs(var.ub)-1)**2
                    F += R*max(0,x[i]-var.ub)**2
            
            return F
        
        return penalty_fun

    def barrier_fun(self,R,barrier_type='log'):
        """ Returns the barrier function
            barrier_type: 'log' or 'inverse'
        """
        
        def bar_fun(x):
            
            F = self.obj(x)
            
            for con in self.cons:
                val = con(x)
                if con.type == '=':
                    """ For equality constraints, the barrier term
                        is not defined
                    """
                    
                elif con.type == '<':
                    """ For less than or equal constraints, the penalty
                        term is -1/R*log(-val)
                    """
                    if barrier_type == 'log':            
                        F -= 1/R*np.log(-val)
                    elif barrier_type == 'inverse':
                        F -= 1/R/val
                elif con.type == '>':
                    """ For geq type constraints, the penalty term is
                        square of min(0,val)
                    """
                    if barrier_type == 'log':            
                        F -= 1/R*np.log(val)
                    elif barrier_type == 'inverse':
                        F += 1/R/val
            
            for (i,var) in enumerate(self.vars):                
                if var.lb > XLB:                    
                    #F += R*max(0,-1-x[i]/abs(var.lb))**2
                    F -= 1/R/(-x[i]+var.lb)
                
                if var.ub < XUB:
                    #F += R*max(0,x[i]/abs(var.ub)-1)**2
                    F -= 1/R/(x[i]-var.ub)        
            
            return F
        
            
        
        return bar_fun

    def augmented_lagrangian(self,Rineq=1.0,Req=1.0):
        """ Constructs the augmented Lagrangian function for the
            problem.
            
            mult .. list of Lagrange multipliers
            R .. penalty factors
            
            NOTE! variable bounds are treated as regular inequalities
        """
        
        def aul_fun(x):
        
            F = self.obj(x)
            
            for con in self.cons:
                val = con(x)
                mult = con.mult
                
                if con.type == '=':
                    """ Equality constraints
                    """
                    F += -mult*val + Req*val**2
                    
                elif con.type == '<':
                    """ Less than or equal constraints
                    """                    
                    F += Rineq*max(0.5*mult/Rineq+val,0)**2
                    
                elif con.type == '>':
                    """ Less than or equal constraints
                        NOTE! CHECK THIS!!!!
                    """                    
                    F += Rineq*max(0.5*mult/Rineq-val,0)**2
          
            for (i,var) in enumerate(self.vars):                     
                if var.lb > XLB:
                    val = -x[i] + var.lb                            
                    F += Rineq*max(0.5*var.lb_mult/Rineq + val,0)**2
                
                if var.ub < XUB:                    
                    val = x[i] - var.ub
                    F += Rineq*max(0.5*var.ub_mult/Rineq + val,0)**2
          
            return F
        
        return aul_fun
    
    def update_lag_mult(self,x,Rineq=1.0,Req=1.0):
        """ Updates Lagrange multipliers according to the update rules
            associated with the augmented Lagrangian method.
        """
        
        """ First, iterate over regular constraints """
        for con in self.cons:
            val = con(x)
            mult = con.mult
            
            if con.type == '=':
                """ Equality constraints
                """
                con.mult -= 2*Req*val
                    
            elif con.type == '<':
                """ Less than or equal constraints
                """                           
                con.mult = max(mult + 2*Rineq*val,0)
                    
            elif con.type == '>':
                """ Less than or equal constraints
                NOTE! CHECK THIS!!!!
                """                    
                con.mult =  max(mult - 2*Rineq*val,0)
         
        """ Iterate over variable bound constraints """
        for (i,var) in enumerate(self.vars):                     
            if var.lb > XLB:
                val = -x[i] + var.lb                            
                var.lb_mult = max(var.lb_mult + 2*Rineq*val,0)
                                
            if var.ub < XUB:                    
                val = x[i] - var.ub
                var.ub_mult = max(var.lb_mult + 2*Rineq*val,0)
            




if __name__ == '__main__':

    b = BinaryVariable(name="oma",section=1,target=3)    

    """    
    prop = OptimizationProblem(name="Testi")

    def obj(x):
        a, b = x
        return a ** 2 + b + 10

    objective = ObjectiveFunction(name="Testi", obj_fun=obj)

    var1 = Variable("Var1", lb=0, ub=100)
    var2 = Variable("Var2", lb=0, ub=100)

    prop.add(objective)
    prop.add(var1)
    prop.add(var2)

    print(prop([5, 5]))

    # dvars = []
    # dvars.append(Variable("Height", 150, 1000))
    # dvars.append(Variable("Flange width", 150, 500))
    # dvars.append(Variable("Flange thickness", 5, 40))
    # dvars.append(Variable("Web thickness", 5, 40))
    #
    """
    """    
    def obj_fun(x):
        return x[0]**2 + 2*x[1]**3 -4*x[2]
    
    x1 = DiscreteVariable(name='x1',values = [4.5, 5.4, 7.23, 11.2, 13.4])
    x2 = IntegerVariable(name='x2',lb=1,ub =10)
    x3 = DiscreteVariable(name='x3',values = [3.77,6.987,11.223])
    x = [x1,x2,x3]
    x0 = [6.66,4.445,5.0]
    
    p = OptimizationProblem(name="I-Beam Weight Minimization",variables=x)
    
    p.obj = obj_fun
    
    N = p.discrete_neighborhood(x0,k=4)
    
    fvals = []
    for xi in N:
        print(xi)
        fvals.append((p.obj(xi)))
    # create grid of neighborhood points
    #M = np.meshgrid(*N)
    
    #s = M[0][0].shape
    """

    """
    for a in np.nditer(M):
        print(len(a),list(a))
    """
    
    """
    for i in range(s[0]):
        for j in range(s[1]):
            for k in 
            for I,m in enumerate(M):
                x[I] = m(k,i,j)
    """         
    
    # p = OptimizationProblem(name="I-Beam Weight Minimization",
    # variables=dvars)
    """
    problem = OptimizationProblem()
    vars = []
    for i in range(10):
        var = Variable("Name " + str(i),
                       lb=0,
                       ub=1)
        vars.append(var)

    problem.vars = vars
    var0 = problem.vars[0]
    var0.lock()
    var3 = problem.vars[3]
    var3.lock(0.2)
    print(len(problem.vars))
    var0.unlock()
    var3.unlock()
    print(len(problem.vars))
    """


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Tue Dec 28 15:03:22 2021

Frame Supports

@author: kmela
"""

# --------------------------- SUPPORT CLASSES ----------------------------
class Support:
    def __init__(self, node, dofs, supp_id=1):
        """
        

        Parameters
        ----------
        node : FrameNode
            Node to be supported.
        dofs : List
            List of supported degrees of freedom.
        supp_id : TYPE, optional
            DESCRIPTION. The default is 1.

        Returns
        -------
        None.

        """
        """ Constructor
            coordinate .. nodal coordinates [list]
            dofs .. list of supported degrees of freedom
            supp_id .. each support has an ID that relates it to a given
                    support system. NOTE! In 'calculate' method, supp_id = 1 is always assumed!
        """
        self.node = node
        self.supp_id = supp_id
        self.dofs = dofs

    def add_support(self, fem_model):
        """ Adds support to the given FEM model """        
        try: 
            idx = fem_model.nodes.index(self.node.fem_node)    
            fem_model.add_support(self.supp_id, idx, self.dofs, val=0.0)
        except ValueError:
            print("Setting of support not successful.")


class FixedSupport(Support):
    '''
    In ROBOT this setting means that UX, UZ and RY check marks are present in the support dialog
    '''
    def __init__(self, node, supp_id=1):
        # super().__init__(coordinate, [1, 1, 1], supp_id)
        super().__init__(node, [0, 1, 2], supp_id)


class XHingedSupport(Support):
    '''
    In ROBOT this setting means that UX direction check marks is present in the support dialog, but UZ and RY are empty
    '''
    def __init__(self, node, supp_id=1):
        super().__init__(node, [0], supp_id)


class YHingedSupport(Support):
    '''
    In ROBOT this setting means that UZ direction check mark is present in the support dialog, but UX and RY are empty
    '''
    def __init__(self, node, supp_id=1):
        super().__init__(node, [1], supp_id)


class XYHingedSupport(Support):
    '''
    In ROBOT this setting means that UX and UZ direction check marks are present in the support dialog, but RY is empty
    '''
    def __init__(self, node, supp_id=1):
        super().__init__(node, [0, 1], supp_id)



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
""" General class for truss structures

    Classes:
        Truss -- general truss structure
        TrussMember -- general truss member
        TrussNode -- general truss node
   
    Written by Kristo Mela 18.4.2017
"""
import math
import numpy as np
import matplotlib.pyplot as plt

class Truss:
    """ General truss structure
        Attributes:
            members -- list of truss members
            nodes -- list of truss nodes
            ID -- table for numbering degrees of freedom
            loads -- array of nodal loads
            loadcases -- array of load case data
            
        Methods:
            add_member -- adds a new member
            add_node -- adds a new node
            nnodes -- get number of nodes
            nmembers -- get number of members
            nlc -- get number of load cases
            member_coord -- get nodal coordinates of a member
            member_length -- compute the length of a member
            member_direction_vector -- determine the direction vector of
                                       a member
            common_node -- determine the common node shared by two members
            angle_between_members -- determine the angle between two members
            members_at_node -- find members connected to a node
        """
    
    
    def __init__(self):
        """ 
        
        x -- 3 x nN matrix of nodal coordinates
        members   % array of truss members
                  % each element is a struct with the following fields:
                  % .n -- 2 x 1 vector with node indices of the members
                  % .profile -- member profile
        ID        % ID table of degrees of freedom
        loads     % array of nodal loads
        loadcases % array of load case data
        """
                
        self.members = {}
        self.nodes = {}
        self.ID = []
        self.load = []
        self.loadcases = []
        self.dim = 2
    
    def add_member(self,n1,n2,profile,label="Member"):
        """ Adds a new member to the truss
            input: 
                n1 .. end node 1
                n2 .. end node 2
                the nodes are "TrussNode" objects
                   
        """
        #new_member = TrussMember([n1,n2],p,type)
        new_member = TrussMember([self.nodes[n1],self.nodes[n2]],profile,label)
        self.members.append(new_member)
    
    def add_node(self,x,y,z=0,label="Node"):
        """ Adds a node to the truss """
        
        """ 
            The dimension of the truss (2D or 3D)
            is determined by the number of coordinates
        """
        if z != 0:
            self.dim = 3
            
        new_node = TrussNode(x,y,z,label)
        self.nodes.append(new_node)
    
    def nnodes(self):
        """ Number of nodes """
        return len(self.nodes)
    
    def nmembers(self):
        """ Number of members """
        return len(self.members)
    
    def nlc(self):
        """ Number of load cases """
        return len(self.loadcases)

    def common_node(self,m1,m2):
        """ Find the common node of members m1 and m2 
            Input:
                m1 .. index of member 1
                m2 .. index of member 2        
        """
        n1 = self.members(m1).nodes
        n2 = self.members(m2).nodes
        if n1[0] in n2:
            ncommon = n1[0]
        else:
            ncommon = n1[1]
            
        return ncommon

    def angle_between_members(self,m1,m2):
        """ Angle (in degrees) between two members 
            Input:
                m1 .. index of member 1
                m2 .. index of member 2        
        """
        
        # it is assumed that the members share a node
        N1 = self.members[m1].nodes
        N2 = self.members[m2].nodes
        
        # find common node
        if N1[0] in N2:
            Ncommon = N1[0]
            # get the nodes of each member that are not shared
            N1unique = N1[1]
            if N2[0] == Ncommon:
                N2unique = N2[1]
            else:
                N2unique = N2[0]
                            
        elif N1[1] in N2:
            Ncommon = N1[1]
            N1unique = N1[0]
            if N2[0] == Ncommon:
                N2unique = N2[1]
            else:
                N2unique = N2[0]
        else:
            Ncommon = None
                                            
        # coordinates of the common node
        Xcommon = self.nodes[Ncommon].x
        # Coordinates of the unique nodes
        X1 = self.nodes[N1unique].x
        X2 = self.nodes[N2unique].x
                    
        v1 = X1-Xcommon
        r1 = v1/np.linalg.norm(v1)
        v2 = X2-Xcommon
        r2 = v2/np.linalg.norm(v2)

        cd = np.dot(r1,r2)
        a = math.degrees(math.acos(cd))
        
        return a
    
    def members_at_node(self,n):
        """ Find members connected to node 'n' """
        nid = id(self.nodes[n])
    
        mems = []
    
        for i in range(self.nmembers()):
            if id(self.members[i].nodes[0]) == nid or \
                id(self.members[i].nodes[1]) == nid:
                mems.append(i)

        return mems
    
    def draw(self):
        """ draw nodes """
        for n in self.nodes:
            plt.plot(n.x[0],n.x[1],'ro')
        
        for i in range(self.nnodes()):
            plt.text(self.nodes[i].x[0],self.nodes[i].x[1],str(i))
        
        """ draw members """
        for i in range(self.nmembers()):
            #X = self.member_coord(i)
            X = self.members[i].coord()
            plt.plot(X[:,0],X[:,1],'b')
            Xmid = X[0,:]+0.5*(X[1,:]-X[0,:])
            plt.text(Xmid[0],Xmid[1],str(i))
        #plt.axis([0, 6, 0, 20])
        
        plt.show()
    
class TrussMember:
    """ Class for truss members.
    
        Basically only for storing data
    """
    
    def __init__(self,n,profile,label="Member"):
        """ Truss member
        
            n -- array of TrussNode objects
            profile -- cross-section of the member
            label -- member label
        """
        self.nodes = n
        self.profile = profile
        self.label = label

    def coord(self):
        """ Nodal coordinates of member m
        
            m -- member index
            xorig -- True: use original nodal locations
                     False: use modified nodal locastions
        """
        X = np.array([self.nodes[0].x,self.nodes[1].x])
        return X

    def length(self):
        X = self.coord()
        L = np.linalg.norm(X[1,:]-X[0,:])
        return L

    def direction_vector(self):
        """ Unit direction vector of member
        
        xorig -- True: use original nodal locations
        False: use modified nodal locastions
        """
        X = self.coord()
        v0 = X[1,:]-X[0,:]
        v = v0/np.linalg.norm(v0)
        return v

class TrussNode:
    """ Class for truss nodes.
    
        Basically only for storing data
    """
    
    def __init__(self,x,y,z=0,label="Node"):
        """ input: x, y and z are the coordinates of the node 
            label .. unique nodel label
        
        """
        self.x = np.array([x,y,z])
        self.label = label
        #self.xmod = np.array([x,y,z])

""" Testing function """
if __name__ == "__main__":
    import sys
    
    from sections.steel.RHS import SHS
    
    t = Truss()
    t.add_node(0,0)
    t.add_node(1,1)
    t.add_node(2,0)
    t.add_node(0,1)
    t.add_node(3,1)
    
    p = SHS(100,5)
    
    t.add_member(0,1,p)
    t.add_member(0,2,p)
    t.add_member(3,1,p)
    t.add_member(2,1,p)
    t.add_member(2,4,p)
    t.add_member(1,4,p)
    
    for m in t.members:
        print(m.coord())
        print(m.length())
    
    print(t.members_at_node(2))
    
    t.draw()



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-

try:
    from metku.framefem.framefem import Material
except:
    from framefem.framefem import Material


class Steel(Material):

    def __init__(self, fy, fu, name):
        super().__init__(210e3, 7850e-9, nu=0.3)
        self.fy = fy
        self.fu = fu
        self.name = name

    def __repr__(self):
        return self.name


MATERIALS = dict(
    S235=Steel(235, 360, "S235"),
    S275=Steel(275, 430, "S275"),
    S355=Steel(355, 510, "S355"),
    S280GD=Steel(280, 360, "S280GD"),
    S420=Steel(420, 500, "S420"),
    S355MC=Steel(355.0, 430.0, "S355MC"),
    S500MC=Steel(500.0, 550.0, "S500MC"),
    S700MC=Steel(700.0, 750.0, "S700MC"),
    S355ML=Steel(355.0, 470.0, "S355ML"),
    S500ML=Steel(500.0, 570.0, "S500ML"),

)

MATERIALS["S700E/F"] = Steel(700.0, 780.0, "S700E/F")





from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Optional

from metku.optimization.variables import Variable, DiscreteVariable, IndexVariable, BinaryVariable, BooleanVariable


class VariableTypeEnum(Enum):
    CONTINUOUS = auto()
    DISCRETE = auto()
    INDEX = auto()
    BINARY = auto()
    BOOLEAN = auto()


def has_attribute(obj: object, attr: list | str) -> bool:
    if isinstance(attr, list):
        for a in attr:
            if not hasattr(obj, a):
                return False
    else:
        return hasattr(obj, attr)
    return True


@dataclass
class VariableGroup:
    name: str
    var_type: VariableTypeEnum
    object: object = None
    objects: list[object] = None
    attribute: Optional[str] = None
    attributes: Optional[list[str]] = None
    value: Optional[list | float] = None
    values: Optional[list] = None
    lower_bounds: Optional[list] = None
    upper_bounds: Optional[list] = None
    variables: list = field(init=False)

    def __post_init__(self):

        if not self.objects:
            self.objects = [self.object]
        if not self.attributes:
            self.attributes = [self.attribute]
        if self.value is None:
            self.value = [0] * len(self.attributes)

        # Attribute check
        for attr in self.attributes:
            for obj in self.objects:
                if not has_attribute(obj, attr):
                    raise TypeError(f"Object {type(obj)} has no attribute: {attr}")

        variables = []
        match self.var_type:

            case VariableTypeEnum.CONTINUOUS:
                for attr, lb, ub, value in zip(self.attributes,
                                               self.lower_bounds,
                                               self.upper_bounds,
                                               self.value):
                    var = Variable(name=f"{self.name} {attr}",
                                   value=value,
                                   lb=lb,
                                   ub=ub,
                                   target={
                                       'property': attr,
                                       'objects': self.objects
                                   })
                    variables.append(var)
            case VariableTypeEnum.DISCRETE:
                for attr, value, values in zip(self.attributes,
                                               self.value,
                                               self.values):
                    var = DiscreteVariable(name=f"{self.name} {attr}",
                                           value=value,
                                           values=values,
                                           target={
                                               'property': attr,
                                               'objects': self.objects
                                           })
                    variables.append(var)
            case VariableTypeEnum.INDEX:
                var = IndexVariable(name=f"{self.name} {self.attribute}",
                                    value=self.value,
                                    values=self.values,
                                    target={
                                        'property': self.attribute,
                                        'objects': self.objects
                                    })
                variables.append(var)
            case VariableTypeEnum.BINARY:
                raise NotImplementedError(f"{VariableTypeEnum.BINARY} Not Implemented")
                # var = BinaryVariable()
            case VariableTypeEnum.BOOLEAN:
                raise NotImplementedError(f"{VariableTypeEnum.BOOLEAN} Not Implemented")
            case _:
                raise ValueError(f"var_type must be VariableTypeEnum, not {self.var_type}")

        self.variables = variables


if __name__ == '__main__':
    from metku.frame2d import SteelBeam
    from metku.sections.steel.catalogue import shs_profiles

    sb = SteelBeam([[0, 0], [5000, 0]])

    vg = VariableGroup(name="ContinuousTest",
                       object=sb.cross_section,
                       attributes=[['Iy', 'Iz'], 'A'],
                       var_type=VariableTypeEnum.CONTINUOUS,
                       # value=[10, 100],
                       lower_bounds=[0, 123],
                       upper_bounds=[1e6, 456])

    vg2 = VariableGroup(name="IndexTest",
                        object=sb,
                        attribute='profile',
                        var_type=VariableTypeEnum.INDEX,
                        values=shs_profiles.keys())

    print(vg.variables)



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
import unittest
from frame2d import *

# Coordinates of members, loads and supports
coord1 = [[0, 0], [0, 3.65e3]]
coord2 = [[0, 3.65e3], [0, 7.3e3]]
coord3 = [[7.3e3, 0], [7.3e3, 3.65e3]]
coord4 = [[7.3e3, 3.65e3], [7.3e3, 7.3e3]]
coord5 = [[0, 3.65e3], [7.3e3, 3.65e3]]
coord6 = [[0, 7.3e3], [7.3e3, 7.3e3]]
supp_coord1 = [0.0, 0]
supp_coord2 = [7.3e3, 0]
# Members
col1 = SteelColumn(coord1, profile='ipe 450')
col2 = SteelColumn(coord2, profile='ipe 400')
col3 = SteelColumn(coord3, profile='ipe 450')
col4 = SteelColumn(coord4, profile='ipe 400')
beam1 = SteelBeam(coord5, profile='ipe 360')
beam2 = SteelBeam(coord6, profile='ipe 300')
# Loads
load1 = PointLoad(coord1[1], [36e3, 0, 0])
load2 = PointLoad(coord2[1], [18e3, 0, 0])
load3 = LineLoad(beam1, [-62.5, -62.5], 'y')
load4 = LineLoad(beam2, [-38, -38], 'y')
# Create empty frame 'envelope'
frame = Frame2D()
# Add members
frame.add(col1)
frame.add(col2)
frame.add(col3)
frame.add(col4)
frame.add(beam1)
frame.add(beam2)
# Add loads
frame.add(load1)
frame.add(load2)
frame.add(load3)
frame.add(load4)
# Add supports
frame.add(FixedSupport(supp_coord1))
frame.add(FixedSupport(supp_coord2))
# Generate frame
frame.generate()
frame.calculate()
# Values from Robot
ALPHA_CR = [4.87840e1, 1.13178e2, 2.53259e2, 4.21790e2]

# Difference allowed between Robot value
DELTA = 0.01


class TestFrame2D(unittest.TestCase):

    def test_add_beam(self):
        frame = Frame2D()
        beam = SteelBeam([[0, 0], [1000, 0]])
        frame.add(beam)
        self.assertEqual(len(frame.beams), 1)

    def test_add_support(self):
        frame = Frame2D()
        sup = FixedSupport([0, 0])
        frame.add(sup)
        self.assertEqual(len(frame.supports), 1)

    def test_add_point_load(self):
        frame = Frame2D()
        pl = PointLoad([0, 0], [-100, 0, 0])
        frame.add(pl)
        self.assertEqual(len(frame.point_loads), 1)

    def test_add_lineload(self):
        frame = Frame2D()
        beam = SteelBeam([[0, 0], [1000, 0]])
        frame.add(LineLoad(beam, [-10, -10], 'y'))
        self.assertEqual(len(frame.line_loads), 1)

    def test_simple_beam(self):
        frame = Frame2D()
        beam = SteelBeam([[0, 0], [1000, 0]])
        frame.add(beam)
        frame.add(LineLoad(beam, [-10, -10], 'y'))
        frame.add(YHingedSupport([0, 0]))
        frame.add(XYHingedSupport([1000, 0]))
        frame.generate()
        frame.calculate()
        self.assertAlmostEqual(beam.med, 1.25e6, places=-3)

    def test_simple_beam2(self):
        frame = Frame2D()
        beam = SteelBeam([[0, 0], [1000, 0]])
        frame.add(beam)
        frame.add(LineLoad(beam, [-10, -10], 'y'))
        frame.add(YHingedSupport([0, 0]))
        frame.add(YHingedSupport([250, 0]))
        frame.add(XYHingedSupport([1000, 0]))
        frame.generate()
        frame.calculate()
        self.assertAlmostEqual(beam.med, -0.547 * 1e6, places=-3)

    def test_simple_beam3(self):
        frame = Frame2D()
        mem = SteelBeam([[0, 0], [5000, 0]])
        mem2 = SteelBeam([[5000, 0], [8000, 0]])
        frame.add(mem)
        frame.add(mem2)
        frame.add(YHingedSupport([8000, 0]))
        frame.add(XYHingedSupport([2000, 0]))
        frame.add(LineLoad(mem2, [-15, -15], 'y'))
        frame.add(PointLoad([0, 0], [0, -20e3, 0]))
        frame.add(PointLoad([4000, 0], [0, 0, -30e6]))
        frame.generate()
        frame.calculate()
        self.assertAlmostEqual(frame.f.elements[6].bending_moment[0], 28.75e6,
                               places=-3)

    #
    # def test_Med(self):
    #      self.assertAlmostEqual(col1.med, -101.14, delta=DELTA)
    #      self.assertAlmostEqual(col2.med, -139.82, delta=DELTA)
    #      self.assertAlmostEqual(col3.med, -175.56, delta=DELTA)
    #      self.assertAlmostEqual(col4.med, -300.62, delta=DELTA)
    #      self.assertAlmostEqual(beam1.med, -300.62, delta=DELTA)
    #      self.assertAlmostEqual(beam2.med, 175.56, delta=DELTA)
    #
    # def test_Ved(self):
    #      self.assertAlmostEqual(col1.ved, -24.11e3, places=-1)
    #      self.assertAlmostEqual(col2.ved, -72.19e3, places=-1)
    #      self.assertAlmostEqual(col3.ved, -143.60e3, places=-1)
    #      self.assertAlmostEqual(col4.ved, -238.51e3, places=-1)
    #      self.assertAlmostEqual(beam1.ved, 78.11e3, places=-1)
    #      self.assertAlmostEqual(beam2.ved, 90.19e3, places=-1)
    #
    # def test_Ned(self):
    #      self.assertAlmostEqual(col1.ned, 351.55e3, places=-1)
    #      self.assertAlmostEqual(col2.ned, 133.80e3, places=-1)
    #      self.assertAlmostEqual(col3.ned, 90.19e3, places=-1)
    #      self.assertAlmostEqual(col4.ned, -12.08e3, places=-1)
    #      self.assertAlmostEqual(beam1.ned, 382.10e3, places=-1)
    #      self.assertAlmostEqual(beam2.ned, 143.60e3, places=-1)

    def test_false_profile(self):
        with self.assertRaises(TypeError):
            beam1.profile = 'iipe 100'

    # def test_alphacr(self):
    #    self.assertAlmostEqual(frame.alpha_cr[0]*1e3, ALPHA_CR[0], delta=DELTA)


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Fri Nov 01 2019

Solving linear programming problems.

@author: kmela
"""

from scipy.optimize import linprog
import numpy as np
import time

# import gurobipy as grb

from metku.optimization.solvers.optsolver import OptSolver
import metku.optimization.structopt as sopt

#from ortools.linear_solver import pywraplp

class LP(OptSolver):

    def __init__(self,algorithm="gurobi"):
        super().__init__()
        self.algo = algorithm
        

    def solve(self,problem,verb=False):
        """
        
        """
    
        if self.algo == "gurobi":
            import gurobipy as grb
            
            try:
                lp = grb.Model("lp")
            except:
                self.algo = 'ortools'
                from ortools.linear_solver import pywraplp
                lp = pywraplp.Solver.CreateSolver('GLOP')
                
        else:
            from ortools.linear_solver import pywraplp
            lp = pywraplp.Solver.CreateSolver('GLOP')
                
        x = []
        for var in problem.vars:
            """
                create variable, including the objective function
                coefficient and name
            """
            if self.algo == "gurobi":
                x.append(lp.addVar(var.lb,var.ub,vtype=grb.GRB.CONTINUOUS,name=var.name))
            else:
                x.append(lp.NumVar(var.lb,var.ub,var.name))
                
        #nvars = len(x)
                
        if problem.obj.obj_type == "MIN":
            if self.algo == "gurobi":
                obj_sense = grb.GRB.MINIMIZE
            else:
                lp.Minimize(lp.Sum(C * X for C, X in zip(problem.obj.c,x)))
        else:
            if self.algo == "gurobi":
                obj_sense = grb.GRB.MAXIMIZE
            else:
                lp.Maximize(lp.Sum(C * X for C, X in zip(problem.obj.c,x)))
        
        if self.algo == "gurobi":
            lp.setObjective(grb.LinExpr(problem.obj.c,x),sense=obj_sense)
            
        # Generate constraints
        for con in problem.cons:
            if isinstance(con,sopt.LinearConstraint):
                if self.algo == "gurobi":
                    if con.type == '<':
                        con_sense = grb.GRB.LESS_EQUAL
                                #lp.addLConstr(LinExpr(con.a, x), grb.GRB.LESS_EQUAL, con.b)
                                #lp.addConstr(grb.quicksum([con.a[j]*x[j] for j in range(nvars)]) <= con.b[i])
                    elif con.type == '>':
                        con_sense = grb.GRB.GREATER_EQUAL
                                #lp.addLConstr(LinExpr(con.a, x), grb.GRB.LESS_EQUAL, con.b)
                                #lp.addConstr(grb.quicksum([con.a[j]*x[j] for j in range(nvars)]) >= con.b[i])
                    else:
                        con_sense = grb.GRB.EQUAL
                                #lp.addConstr(grb.quicksum([con.a[j]*x[j] for j in range(nvars)]) = con.b[i])
                    
                    lp.addLConstr(grb.LinExpr(con.a, x), con_sense, con.b)
                else:
                    lhs = lp.Sum(A * X for A, X in zip(con.a,x))
            
                    if con.type == '<':
                        lp.Add(lhs <= con.b)
                    elif con.type == '>':
                        lp.Add(lhs >= con.b)
                    else:
                        lp.Add(lhs == con.b)
        
        if self.algo == "gurobi":
            lp.update()                
       

            #for var in lp.getVars():
            #    print(var.lb,var.ub)
            
            
            if not verb:
                lp.setParam("LogToConsole",0)
                    #qp.update()
            lp.optimize()
                    
            if lp.Status == grb.GRB.OPTIMAL:
                self.feasible = True
            elif lp.Status == grb.GRB.INFEASIBLE:
                self.feasible = False
                    
            X = []
            for v in lp.getVars():
                #print('%s %g' % (v.varName, v.x))                    
                X.append(v.x)
            
            # bounds = [x*self.move_limits for x in self.X]
            #
            # res = linprog(df, A, B, bounds=bounds)
        else:
            status = lp.Solve()
            
            X = []
            if status == lp.OPTIMAL:
                for v in lp.variables():
                    X.append(v.solution_value())    

        self.X = X
        #return X

def Arora_Ex62():
    x1 = sopt.Variable('x1',0,1e4)
    x2 = sopt.Variable('x2',0,1e4)
    p = sopt.OptimizationProblem(name="Arora, Example 6.2", variables=[x1,x2])
    
    c = [-400,-600]
    
    obj = sopt.LinearObjective("Objective",c,problem=p)
    p.add(obj)
    
    cons = []
    cons.append(sopt.LinearConstraint([1,1],16,name="Con1",problem=p))
    cons.append(sopt.LinearConstraint([1/28,1/14],1,name="Con2",problem=p))
    cons.append(sopt.LinearConstraint([1/14,1/24],1,name="Con3",problem=p))
    
    for con in cons:
        p.add(con)
    
    solver = LP("gurobi")
    solver.solve(p)
    p(solver.X)
    
def Arora_Ex63():
    x1 = sopt.Variable('x1',0,1e4)
    x2 = sopt.Variable('x2',0,1e4)
    p = sopt.OptimizationProblem(name="Arora, Example 6.3", variables=[x1,x2])
    
    c = [4,5]
    
    obj = sopt.LinearObjective("Objective",c,obj_type="MAX",problem=p)
    p.add(obj)
    
    cons = []
    cons.append(sopt.LinearConstraint([-1,1],4,name="Con1",problem=p))
    cons.append(sopt.LinearConstraint([1,1],6,name="Con2",problem=p))

    
    for con in cons:
        p.add(con)
    
    #solver = LP("gurobi")
    solver = LP("ortools")
    solver.solve(p)
    p(solver.X)

    return solver, p

def Arora_Ex67():
    x = []
    for i in range(4):
        x.append(sopt.Variable('x'+str(i+1),0,1e4))
            
    p = sopt.OptimizationProblem(name="Arora, Example 6.7", variables=x)
    
    c = [4,5,0,0]
    
    obj = sopt.LinearObjective("Objective",c,obj_type="MAX",problem=p)
    p.add(obj)
    
    cons = []
    cons.append(sopt.LinearConstraint([-1,1,1,0],4,name="Con1",con_type='=',problem=p))
    cons.append(sopt.LinearConstraint([1,1,0,1],6,name="Con2",con_type='=',problem=p))

    
    for con in cons:
        p.add(con)
    
    #solver = LP("gurobi")
    solver = LP("ortools")
    solver.solve(p)
    p(solver.X)
    
    return solver, p

if __name__ == '__main__':
    
    solver, p = Arora_Ex63()
    

    
    
    
    




# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Timber section parameters, including timber fire design calculations (simple method)

@author: Viktor Haimi
"""

from abc import ABC, abstractmethod
import numpy as np
from scipy import integrate

from metku.eurocodes.en1995 import en1995_1_1, en1995_1_2
from metku.eurocodes.en1995.fire_protection import *
from metku.materials.timber_data import Timber, T

import matplotlib.pyplot as plt
import math

import sys

if sys.version_info[0] == 3 and sys.version_info[1] < 9:
    from typing import List
    list_out = List[float]
else:
    list_out = list[float]


class TimberSection:
    def __init__(self, B: (int, float), H: (int, float), material: T=T.GL30c,
                 fire_protection_right: FireProtection = None, fire_protection_left: FireProtection = None,
                 fire_protection_top: FireProtection = None, fire_protection_bottom: FireProtection = None,
                 fire_protection_generic: FireProtection = None, fire_protection_sides: int = 3,
                 sides_on_fire: int = 3, lvldir: str = 'edge'):
        """
        Rectangular cross-section for a prismatic beam
        
        @param B: width [mm]
        @param H: depth (height) [mm]
        @param material: material
        @param fire_protection_right: fire protection method for right-hand-side, overrides fire_protection_genericin
        @param fire_protection_left: fire protection method for left-hand-side, overrides fire_protection_genericin
        @param fire_protection_top: fire protection method for top edge, overrides fire_protection_genericin
        @param fire_protection_bottom: fire protection method for bottom edge, overrides fire_protection_genericin
        @param fire_protection_generic: generic fire protection that is given for several sides at once
        @param fire_protection_sides: = 3 means that other sides except the top edge are protected 'fire_protection_top'.
                                      != 3 means that all four sides are fire protected with 'fire_protection_generic'
        @param sides_on_fire: = 4 means that fire attacks the section from four sides,
                              != 4 means that fire attacks the section from all other sides, except top
        @param lvldir: orientation of lvl. 'edge' (Finn. lappeellaan) or 'flat' ('Finn. syrjälllään')
        """
        self._H = H
        self._B = B

        self.timber_member = None

        self.beta_0 = None
        self.beta_n = None # Charring speed
        
        self.R = 0 # Fire duration

        self.__fire_protection_generic = fire_protection_generic
        self.fire_protection_sides = fire_protection_sides
        self.sides_on_fire = sides_on_fire
        if fire_protection_sides == 3:
            self.fire_protection_right = fire_protection_right if fire_protection_right else fire_protection_generic
            self.fire_protection_left = fire_protection_left if fire_protection_left else fire_protection_generic
            self.fire_protection_bottom = fire_protection_bottom if fire_protection_bottom else fire_protection_generic
            self.fire_protection_top = fire_protection_top
        else:
            self.fire_protection_right = fire_protection_right if fire_protection_right else fire_protection_generic
            self.fire_protection_left = fire_protection_left if fire_protection_left else fire_protection_generic
            self.fire_protection_bottom = fire_protection_bottom if fire_protection_bottom else fire_protection_generic
            self.fire_protection_top = fire_protection_top if fire_protection_top else fire_protection_generic

        self.lvldir = lvldir
        self.material = material

    @property
    def material(self) -> Timber:
        return self.__material

    def set_beta_0_and_n(self):
        """
        laskee ja asettaa hiiltymisnopeudet beta_0:n ja beta_n:n
        SFS-EN1995-1-2 Taulukko 3.1 Tai RIL-205-2-2009 Taulukon 3.2
        @return: None
        """
        self.beta_0, self.beta_n = en1995_1_2.charring_speed(self.material.rhok, self.material.type,
                                                             self.material.hardness, min(self._B, self._H))

    @material.setter
    def material(self, val: T):
        self.__material = Timber(val)
        if self.__material.type == 'lvl':
            self.__material.direction = self.lvldir
        self.set_beta_0_and_n()

        # RIL-205-2-2009 3.4.3
        # Asetetaan palosuojille millainen materiaali suojataan jotta voidaan laskea ta ja tf
        if self.fire_protection_right:
            self.fire_protection_right.protected_material(self.material)
            if isinstance(self.fire_protection_right, (GypsumPlasterboardF, GypsumPlasterboardFF, GypsumPlasterboardAF)):
                self.fire_protection_right.set_tf_for_gypsumF()
            self.fire_protection_right.set_ta()
        if self.fire_protection_left:
            self.fire_protection_left.protected_material(self.material)
            if isinstance(self.fire_protection_left, (GypsumPlasterboardF, GypsumPlasterboardFF, GypsumPlasterboardAF)):
                self.fire_protection_left.set_tf_for_gypsumF()
            self.fire_protection_left.set_ta()
        if self.fire_protection_bottom:
            self.fire_protection_bottom.protected_material(self.material)
            if isinstance(self.fire_protection_bottom, (GypsumPlasterboardF, GypsumPlasterboardFF, GypsumPlasterboardAF)):
                self.fire_protection_bottom.set_tf_for_gypsumF()
            self.fire_protection_bottom.set_ta()
        if self.fire_protection_top:
            self.fire_protection_top.protected_material(self.material)
            if isinstance(self.fire_protection_top, (GypsumPlasterboardF, GypsumPlasterboardFF, GypsumPlasterboardAF)):
                self.fire_protection_top.set_tf_for_gypsumF()
            self.fire_protection_top.set_ta()

    @property
    def H(self) -> float:
        h = self._H
        h -= en1995_1_2.d_char_n(self.fire_protection_bottom, self.R, self.beta_n)
        if self.sides_on_fire == 4:
            h -= en1995_1_2.d_char_n(self.fire_protection_top, self.R, self.beta_n)
        if h <= 0:
            h = 1e-5
        return h

    @H.setter
    def H(self, val: (int, float)):
        self._H = val

    def get_H(self, x: (int, float)) -> (int, float):
        return self.H

    def get_A(self, x: (int, float)) -> (int, float):
        return self.A

    @property
    def B(self) -> float:
        b = self._B
        b -= en1995_1_2.d_char_n(self.fire_protection_right, self.R, self.beta_n)
        b -= en1995_1_2.d_char_n(self.fire_protection_left, self.R, self.beta_n)
        if b <= 0:
            b = 1e-5
        return b

    @B.setter
    def B(self, val: (int, float)):
        self._B = val

    @property
    def I(self) -> np.ndarray:
        return np.asarray([(self.B * self.H ** 3) / 12, (self.H * self.B ** 3) / 12])

    def get_I(self, x: (int, float)) -> np.ndarray:
        return self.I

    @property
    def Wel(self) -> np.ndarray:
        return np.asarray([(self.B * self.H ** 2) / 6, (self.H * self.B ** 2) / 6])

    def get_Wel(self, x: (int, float)) -> np.ndarray:
        return self.Wel

    @property
    def k_h(self) -> float:
        # SFS-EN1995-1-1 3.2, 3.3, 3.4
        # TODO pitää kysyä samilta
        # toimii kaikille rajatiloille, jos halutaan poistaa palolta niin pitää muokata koodia
        Med = max(self.timber_member.MEdY, self.timber_member.MEdZ)
        if self.material.rhok <= 700 and self.material.type == 'solid_timber':
            if self.timber_member.NEd > 0 and self.B < 150 and Med != 0 and self.H < 150:
                return min(min(1.3, (150 / self.B) ** 0.2), min(1.3, (150 / self.H) ** 0.2))
            elif self.timber_member.NEd > 0 and self.B < 150:
                return min(1.3, (150 / self.B) ** 0.2)
            elif Med != 0 and self.H < 150:
                return min(1.3, (150 / self.H) ** 0.2)
            else:
                return 1.0
        elif self.material.type == 'glt':
            if self.timber_member.NEd > 0 and self.B < 600 and Med != 0 and self.H < 600:
                return min(min(1.1, (600 / self.B) ** 0.1), min(1.1, (600 / self.H) ** 0.1))
            elif self.timber_member.NEd > 0 and self.B < 600:
                return min(1.1, (600 / self.B) ** 0.1)
            elif Med != 0 and self.H < 600:
                return min(1.1, (600 / self.H) ** 0.1)
            else:
                return 1.0
        elif self.material.type == 'lvl':
            if Med != 0 and self.H != 300:
                return min(1.2, (300 / self.H) ** self.material.s)
            else:
                return 1.0
        else:
            return 1.0

    @property
    def A(self) -> (int, float):
        '''
        RIL-205-2-2009 4.1.1S 4.1.2S
        Laskee palolle altistuneen rakenteen poikkileikkauksen tehollisen pinta-alan palonkestoajan funktiona
        @return: tehollinen pinta-ala A_ef [mm^2]
        '''

        b = self.B
        h = self.H

        if b <= 0 or h <= 0:
            return 0
        return b*h

    @property
    def fire_protection_generic(self) -> FireProtection:
        return self.__fire_protection_generic

    @fire_protection_generic.setter
    def fire_protection_generic(self, val: FireProtection):
        self.__fire_protection_generic = val
        if self.fire_protection_sides == 3:
            self.fire_protection_right = val
            self.fire_protection_left = val
            self.fire_protection_bottom = val
        else:
            self.fire_protection_right = val
            self.fire_protection_left = val
            self.fire_protection_bottom = val
            self.fire_protection_top = val

    def get_R(self) -> (int, float):
        return self.R

    def __str__(self):
        return f'TimberSection({round(self._B, 0)}X{round(self._H, 0)})'

    def __repr__(self):
        return f"TimberSection {self.material.timber_type} {round(self.H, 0)}X{round(self.B, 0)}"


class TaperedSection(TimberSection, ABC):
    def __init__(self, B: (int, float), H: (int, float), material: T=T.C14,
                 fire_protection_right: FireProtection = None, fire_protection_left: FireProtection = None,
                 fire_protection_top: FireProtection = None, fire_protection_bottom: FireProtection = None,
                 fire_protection_generic: FireProtection = None, fire_protection_sides=3, sides_on_fire=3, lvldir='edge'):
        super().__init__(B, H, material, fire_protection_right, fire_protection_left,
                         fire_protection_top, fire_protection_bottom, fire_protection_generic, fire_protection_sides,
                         sides_on_fire, lvldir)
        self.small_elements = []
        self.num_elements = -1

    def create_elements(self):
        # TODO ei laske oiken jos nodet ei ole tasavälein
        if len(self.small_elements) > 0:
            raise Exception('Use update_elements to update')
        if self.num_elements < 0:
            raise Exception('num_elements not updated')
        elem_l = 1 / self.num_elements
        for i in range(self.num_elements):
            a = self.get_A(i * elem_l + 0.5 * elem_l)
            iy = self.get_I(i * elem_l + 0.5 * elem_l)
            self.small_elements.append(SmallElementSection(a, iy))

    def update_elements(self):
        elem_l = 1 / self.num_elements
        for i in range(len(self.small_elements)):
            a = self.get_A(i * elem_l + 0.5 * elem_l)
            iy = self.get_I(i * elem_l + 0.5 * elem_l)
            self.small_elements[i].A = a
            self.small_elements[i].I = iy

    def get_H(self, x: (int, float)) -> float:
        """
        Getter for H value

        @param x: local coordinate 0 ... 1
        @return: height at location
        """

        h = self.get_unmodified_h(x)

        h -= en1995_1_2.d_char_n(self.fire_protection_bottom, self.R, self.beta_n)
        if self.sides_on_fire == 4:
            h -= en1995_1_2.d_char_n(self.fire_protection_top, self.R, self.beta_n)
        if h <= 0:
            h = 1e-5
        return h

    def get_A(self, x: (int, float)) -> float:
        """
        Calculates Area of cross section

        @param x: local coordinate 0 ... 1
        @return: area at location
        """
        b = self.B
        h = self.get_H(x)

        if b <= 0 or h <= 0:
            return 1e-5
        return b * h

    def get_Wel(self, x: (int, float)) -> np.ndarray:
        return np.asarray([(self.B * self.get_H(x) ** 2) / 6, (self.get_H(x) * self.B ** 2) / 6])

    def get_I(self, x: (int, float)) -> np.ndarray:
        return np.asarray([(self.B * self.get_H(x) ** 3) / 12, (self.get_H(x) * self.B ** 3) / 12])

    @abstractmethod
    def get_unmodified_h(self, x: (int, float)) -> float:
        pass


class PulpettiPalkki(TaperedSection):
    def __init__(self, B, H0, Hap, material=T.C14, flipped=False, fire_protection_right: FireProtection = None,
                 fire_protection_left: FireProtection = None, fire_protection_top: FireProtection = None,
                 fire_protection_bottom: FireProtection = None, fire_protection_generic: FireProtection = None,
                 fire_protection_sides=3, sides_on_fire=3, lvldir='edge'):
        """
        Pulpettipalkki

        @param B: paksuus [mm]
        @param H0: palkin alkupään korkeus [mm]
        @param Hap: palkin loppupään korkeus[mm]
        @param material: materiaali
        @param flipped: ylösalaisin
        @param fire_protection_right: palosuoja joka laitetaan oikealle, yliajaa fire_protection_genericin
        @param fire_protection_left: palosuoja joka laitetaan vasemmalle, yliajaa fire_protection_genericin
        @param fire_protection_top: palosuoja joka laitetaan ylös, yliajaa fire_protection_genericin
        @param fire_protection_bottom: palosuoja joka laitetaan alas, yliajaa fire_protection_genericin
        @param fire_protection_generic: geneerinen palosuoja joka annetaan monelle sivulle samaan aikaan
        @param fire_protection_sides: jos antaa 3 niin fire_protection_top jää tyhjäksi.
                                      jos antaa jonkun muun kuin 3 niin suojaa kaikki neljä sivua fire_protection_genericillä
        @param sides_on_fire: jos antaa 4 niin poltetaan kaikilta neljältä sivulta,
                              jos antaa jonkun muun kuin 4 niin poltetaan kolmelta sivulta. (ylhäältä ei polteta)
        @param lvldir: lvl:n suuntaus. lappeellaan tai syrjällään. 'edge' tai 'flat'
        """
        self.__material = None
        self.h0 = H0
        self.hap = Hap
        self.flipped = flipped
        super().__init__(B, H0, material=material, fire_protection_right=fire_protection_right,
                         fire_protection_left=fire_protection_left,
                         fire_protection_top=fire_protection_top,
                         fire_protection_bottom=fire_protection_bottom,
                         fire_protection_generic=fire_protection_generic,
                         fire_protection_sides=fire_protection_sides,
                         sides_on_fire=sides_on_fire, lvldir=lvldir)

    @property
    def H0(self) -> float:
        return self.h0

    @property
    def Hap(self) -> float:
        return self.hap

    @H0.setter
    def H0(self, val):
        self.h0 = val
        self.update_elements()

    @Hap.setter
    def Hap(self, val):
        self.hap = val
        self.update_elements()

    def get_unmodified_h(self, x: (int, float)) -> float:
        k = self.hap - self.h0
        h = self.h0 + k * x
        return h

    def __str__(self):
        return f'Pulpettipalkki({round(self._B, 0)}X{round(self.H0, 0)}X{round(self.Hap, 0)})'


class HarjaPalkki(TaperedSection):
    def __init__(self, B, H0, Hap, xap=0.5, material=T.C14, fire_protection_right: FireProtection = None,
                 fire_protection_left: FireProtection = None, fire_protection_top: FireProtection = None,
                 fire_protection_bottom: FireProtection = None, fire_protection_generic: FireProtection = None,
                 fire_protection_sides=3, sides_on_fire=3, lvldir='edge'):
        """
        Harhjapalkki

        @param B: paksuus [mm]
        @param H0: palkin alkupään korkeus [mm]
        @param Hap:
        @param xap:
        @param material:
        @param fire_protection_right: palosuoja joka laitetaan oikealle, yliajaa fire_protection_genericin
        @param fire_protection_left: palosuoja joka laitetaan vasemmalle, yliajaa fire_protection_genericin
        @param fire_protection_top: palosuoja joka laitetaan ylös, yliajaa fire_protection_genericin
        @param fire_protection_bottom: palosuoja joka laitetaan alas, yliajaa fire_protection_genericin
        @param fire_protection_generic: geneerinen palosuoja joka annetaan monelle sivulle samaan aikaan
        @param fire_protection_sides: jos antaa 3 niin fire_protection_top jää tyhjäksi.
                                      jos antaa jonkun muun kuin 3 niin suojaa kaikki neljä sivua fire_protection_genericillä
        @param sides_on_fire: jos antaa 4 niin poltetaan kaikilta neljältä sivulta,
                              jos antaa jonkun muun kuin 4 niin poltetaan kolmelta sivulta. (ylhäältä ei polteta)
        @param lvldir: lvl:n suuntaus. lappeellaan tai syrjällään. 'edge' tai 'flat'
        """
        self.__material = None
        self.h0 = H0
        self.hap = Hap
        self.xap = xap
        super().__init__(B, H0, material=material, fire_protection_right=fire_protection_right,
                         fire_protection_left=fire_protection_left,
                         fire_protection_top=fire_protection_top,
                         fire_protection_bottom=fire_protection_bottom,
                         fire_protection_generic=fire_protection_generic,
                         fire_protection_sides=fire_protection_sides,
                         sides_on_fire=sides_on_fire, lvldir=lvldir)

    def get_unmodified_h(self, x: (int, float)) -> float:
        """
        Defines harjapalkki's height
        @param x: Harjapalkki's longitudinal coordinate
        @param h0: Section height at the beginning and at the end of the beam
        @param hap: Section height at the paramount point of the beam
        @param xap: X coordinate of the highest cross section of the beam
        @param L: Beam span length (support node to support node)
        @return: Height
        """

        h = (((self.hap - self.h0) / self.xap) * x + self.h0 - max(
            ((self.hap - self.h0) / self.xap) * x + self.h0 - ((self.h0 - self.hap) / (1 - self.xap) * x + self.hap +
                                                               ((self.hap - self.h0) / (1 - self.xap)) * self.xap), 0))
        return h

    @property
    def H0(self) -> float:
        return self.h0

    @property
    def Hap(self) -> float:
        return self.hap

    @property
    def Xap(self) -> float:
        return self.xap

    @H0.setter
    def H0(self, val: (int, float)):
        self.h0 = val
        self.update_elements()

    @Hap.setter
    def Hap(self, val: (int, float)):
        self.hap = val
        self.update_elements()

    @Xap.setter
    def Xap(self, val: (int, float)):
        self.xap = val
        self.update_elements()

    def __str__(self):
        return f'Harjapalkki({round(self._B, 0)}X{round(self.h0, 0)}X{round(self.hap, 0)})'


class MahaPalkki(TaperedSection):
    def __init__(self, B, H0, alpha, xap=0.5, material=T.C14, fire_protection_right: FireProtection = None,
                 fire_protection_left: FireProtection = None, fire_protection_top: FireProtection = None,
                 fire_protection_bottom: FireProtection = None, fire_protection_generic: FireProtection = None,
                 fire_protection_sides=3, sides_on_fire=3, lvldir='edge'):
        """
        Mahapalkki harjapalkkitapainen

        @param B: paksuus [mm]
        @param H0: palkin alkupään korkeus [mm]
        @param alpha:
        @param xap:
        @param material:
        @param fire_protection_right: palosuoja joka laitetaan oikealle, yliajaa fire_protection_genericin
        @param fire_protection_left: palosuoja joka laitetaan vasemmalle, yliajaa fire_protection_genericin
        @param fire_protection_top: palosuoja joka laitetaan ylös, yliajaa fire_protection_genericin
        @param fire_protection_bottom: palosuoja joka laitetaan alas, yliajaa fire_protection_genericin
        @param fire_protection_generic: geneerinen palosuoja joka annetaan monelle sivulle samaan aikaan
        @param fire_protection_sides: jos antaa 3 niin fire_protection_top jää tyhjäksi.
                                      jos antaa jonkun muun kuin 3 niin suojaa kaikki neljä sivua fire_protection_genericillä
        @param sides_on_fire: jos antaa 4 niin poltetaan kaikilta neljältä sivulta,
                              jos antaa jonkun muun kuin 4 niin poltetaan kolmelta sivulta. (ylhäältä ei polteta)
        @param lvldir: lvl:n suuntaus. lappeellaan tai syrjällään. 'edge' tai 'flat'
        """
        self.__material = None
        self.h0 = H0
        self.alpha = math.radians(alpha)
        self.beta = None
        self.parent = None
        self.xap = xap
        super().__init__(B, H0, material=material, fire_protection_right=fire_protection_right,
                         fire_protection_left=fire_protection_left,
                         fire_protection_top=fire_protection_top,
                         fire_protection_bottom=fire_protection_bottom,
                         fire_protection_generic=fire_protection_generic,
                         fire_protection_sides=fire_protection_sides,
                         sides_on_fire=sides_on_fire, lvldir=lvldir)

    def get_unmodified_h(self, x: (int, float)) -> float:
        x *= self.parent.length

        ya = self.parent.length * math.tan(self.alpha)
        self.beta = np.arctan(ya / (self.parent.length - self.xap * self.parent.length))

        h = ((math.tan(self.alpha) * x + self.h0) / math.cos(self.alpha) -
             max(0, (x * math.tan(self.beta) - self.xap * self.parent.length * math.tan(self.beta))) / math.cos(self.alpha))
        return h

    @property
    def H0(self) -> float:
        return self.h0

    @property
    def Hap(self) -> float:
        return self.get_H(self.xap)

    @property
    def hap(self) -> float:
        return self.Hap

    @property
    def Xap(self) -> float:
        return self.xap

    @H0.setter
    def H0(self, val: (int, float)):
        self.h0 = val
        self.update_elements()

    @Xap.setter
    def Xap(self, val: (int, float)):
        self.xap = val
        self.update_elements()

    def __str__(self):
        return f'MahaPalkki({round(self._B, 0)}X{round(self.h0, 0)}X{round(self.hap, 0)})'


class KaarevaPalkki(TaperedSection):
    def __init__(self, B, H0, alpha, beta, r_in, material=T.C14, t=45, fire_protection_right: FireProtection = None,
                 fire_protection_left: FireProtection = None, fire_protection_top: FireProtection=None,
                 fire_protection_bottom: FireProtection = None, fire_protection_generic: FireProtection=None,
                 fire_protection_sides=3, sides_on_fire=3, lvldir='edge'):
        """
        Kaarevapalkki

        @param B: paksuus [mm]
        @param H0: palkin alkupään korkeus [mm]
        @param alpha:
        @param beta:
        @param r_in:
        @param material:
        @param t:
        @param fire_protection_right: palosuoja joka laitetaan oikealle, yliajaa fire_protection_genericin
        @param fire_protection_left: palosuoja joka laitetaan vasemmalle, yliajaa fire_protection_genericin
        @param fire_protection_top: palosuoja joka laitetaan ylös, yliajaa fire_protection_genericin
        @param fire_protection_bottom: palosuoja joka laitetaan alas, yliajaa fire_protection_genericin
        @param fire_protection_generic: geneerinen palosuoja joka annetaan monelle sivulle samaan aikaan
        @param fire_protection_sides: jos antaa 3 niin fire_protection_top jää tyhjäksi.
                                      jos antaa jonkun muun kuin 3 niin suojaa kaikki neljä sivua fire_protection_genericillä
        @param sides_on_fire: jos antaa 4 niin poltetaan kaikilta neljältä sivulta,
                              jos antaa jonkun muun kuin 4 niin poltetaan kolmelta sivulta. (ylhäältä ei polteta)
        @param lvldir: lvl:n suuntaus. lappeellaan tai syrjällään. 'edge' tai 'flat'
        """
        self.h0 = H0
        self.parent = None
        self.alpha = np.pi * alpha / 180
        self.beta = np.pi * beta / 180
        self.r_in = r_in
        self.t = t

        super().__init__(B, H0, material=material, fire_protection_right=fire_protection_right,
                         fire_protection_left=fire_protection_left,
                         fire_protection_top=fire_protection_top,
                         fire_protection_bottom=fire_protection_bottom,
                         fire_protection_generic=fire_protection_generic,
                         fire_protection_sides=fire_protection_sides,
                         sides_on_fire=sides_on_fire, lvldir=lvldir)

        self.xap = 0.5
        self.hap = None

    def update_circles(self):
        self.L_y = 0.5 * self.parent.length * np.tan(self.beta)
        self.U_y = 0.5 * self.parent.length * (np.tan(self.alpha) + np.tan(self.beta)) - self.L_y + self.h0

        self.x_circle = self.parent.length / 2
        self.y_circle_up = -self.r_in * np.cos(self.alpha) + (0.5 * self.parent.length - self.r_in * np.sin(self.alpha)) * np.tan(
            self.alpha) + self.h0
        self.y_circle_dw = (0.5 * self.parent.length - self.r_in * np.sin(self.beta)) * np.tan(self.beta) - self.r_in * np.cos(
            self.beta)

        self.up_x1, self.up_y1 = [self.parent.length / 2 - self.r_in * np.sin(self.alpha),
                                  (self.parent.length / 2 - self.r_in * np.sin(self.alpha)) * np.tan(self.alpha) + self.h0]
        self.dwn_x1, self.dwn_y1 = [self.parent.length / 2 - self.r_in * np.sin(self.beta),
                                    (self.parent.length / 2 - self.r_in * np.sin(self.beta)) * np.tan(self.beta)]
        self.up_x2, self.up_y2 = [self.parent.length / 2 - self.r_in * np.sin(self.alpha) + 2 * (self.r_in * np.sin(self.alpha)),
                                  (self.parent.length / 2 - self.r_in * np.sin(self.alpha)) * np.tan(self.alpha) + self.h0]
        self.dwn_x2, self.dwn_y2 = [self.parent.length / 2 - self.r_in * np.sin(self.beta) + 2 * (self.r_in * np.sin(self.beta)),
                                    (self.parent.length / 2 - self.r_in * np.sin(self.beta)) * np.tan(self.beta)]
        self.hap = self.get_H(0.5)

    def f1(self, x: float) -> float:
        return np.tan(self.alpha) * x + self.h0

    def f2(self, x: float) -> float:
        return np.tan(self.beta) * x

    def f3(self, x: float) -> float:
        return (self.parent.length - x) * np.tan(self.alpha) + self.h0

    def f4(self, x: float) -> float:
        return 2 * self.L_y - np.tan(self.beta) * x

    def f5(self, x: float) -> float:
        return np.sqrt(self.r_in ** 2 - (self.parent.length / 2 - x) ** 2) + self.y_circle_up

    def f6(self, x: float) -> float:
        return np.sqrt(self.r_in ** 2 - (self.parent.length / 2 - x) ** 2) + self.y_circle_dw

    @property
    def H0(self) -> float:
        return self.h0

    @property
    def Hap(self) -> float:
        return self.hap

    @H0.setter
    def H0(self, val: (int, float)):
        self.h0 = val
        self.update_circles()
        self.update_elements()

    @property
    def Alpha(self) -> float:
        return self.alpha

    @Alpha.setter
    def Alpha(self, val: (int, float)):
        self.alpha = np.pi * val / 180
        self.update_circles()
        self.update_elements()

    @property
    def Beta(self) -> float:
        return self.beta

    @Beta.setter
    def Beta(self, val: (int, float)):
        self.beta = np.pi * val / 180
        self.update_circles()
        self.update_elements()

    @property
    def R_in(self) -> float:
        return self.r_in

    @R_in.setter
    def R_in(self, val: (int, float)):
        self.r_in = val
        self.update_circles()
        self.update_elements()

    def get_unmodified_h(self, x: (int, float)) -> float:
        x *= self.parent.length
        if 0 <= x < self.up_x1:
            h = self.f1(x) - self.f2(x)
        elif self.up_x1 <= x < self.dwn_x1:
            h = self.f5(x) - self.f2(x)
        elif self.dwn_x1 <= x < self.dwn_x2:
            h = self.f5(x) - self.f6(x)
        elif self.dwn_x2 <= x < self.up_x2:
            h = self.f5(x) - self.f4(x)
        else:
            h = self.f3(x) - self.f4(x)
        return h

    def A_face(self):
        if1 = integrate.quad(self.f1, 0, self.up_x1)
        if2 = integrate.quad(self.f2, 0, self.dwn_x1)
        if3 = integrate.quad(self.f3, self.up_x2, self.parent.length)
        if4 = integrate.quad(self.f4, self.dwn_x2, self.parent.length)
        if5 = integrate.quad(self.f5, self.up_x1, self.up_x2)
        if6 = integrate.quad(self.f6, self.dwn_x1, self.dwn_x2)
        a = if1[0] + if3[0] + if5[0] - if2[0] - if4[0] - if6[0]
        return a

    def __str__(self):
        return f'KaarevaPalkki({round(self._B, 0)}X{round(self.H0, 0)}X{round(self.Alpha, 0)}X{round(self.Beta, 0)}X{round(self.R_in, 0)})'


class KaarevaHarjaPalkki(TaperedSection):
    def __init__(self, B, H0, alpha, beta, r_in, material=T.C14, t=45, fire_protection_right: FireProtection=None, fire_protection_left: FireProtection=None,
                 fire_protection_top: FireProtection=None, fire_protection_bottom: FireProtection=None,
                 fire_protection_generic: FireProtection=None, fire_protection_sides=3, sides_on_fire=3, lvldir='edge'):
        """
        Kaarevaharjapalkki

        @param B: paksuus [mm]
        @param H0: palkin alkupään korkeus [mm]
        @param alpha:
        @param beta:
        @param r_in:
        @param material:
        @param t:
        @param fire_protection_right: palosuoja joka laitetaan oikealle, yliajaa fire_protection_genericin
        @param fire_protection_left: palosuoja joka laitetaan vasemmalle, yliajaa fire_protection_genericin
        @param fire_protection_top: palosuoja joka laitetaan ylös, yliajaa fire_protection_genericin
        @param fire_protection_bottom: palosuoja joka laitetaan alas, yliajaa fire_protection_genericin
        @param fire_protection_generic: geneerinen palosuoja joka annetaan monelle sivulle samaan aikaan
        @param fire_protection_sides: jos antaa 3 niin fire_protection_top jää tyhjäksi.
                                      jos antaa jonkun muun kuin 3 niin suojaa kaikki neljä sivua fire_protection_genericillä
        @param sides_on_fire: jos antaa 4 niin poltetaan kaikilta neljältä sivulta,
                              jos antaa jonkun muun kuin 4 niin poltetaan kolmelta sivulta. (ylhäältä ei polteta)
        @param lvldir: lvl:n suuntaus. lappeellaan tai syrjällään. 'edge' tai 'flat'
        """
        self.h0 = H0
        self.parent = None
        self.alpha = np.pi * alpha / 180
        self.beta = np.pi * beta / 180
        self.r_in = r_in
        self.t = t
        super().__init__(B, H0, material=material, fire_protection_right=fire_protection_right,
                         fire_protection_left=fire_protection_left,
                         fire_protection_top=fire_protection_top,
                         fire_protection_bottom=fire_protection_bottom,
                         fire_protection_generic=fire_protection_generic,
                         fire_protection_sides=fire_protection_sides,
                         sides_on_fire=sides_on_fire, lvldir=lvldir)
        self.hap = None
        self.xap = 0.5

    def update_circles(self):
        self.L_y = 0.5 * self.parent.length * np.tan(self.beta)
        self.U_y = 0.5 * self.parent.length * (np.tan(self.alpha) + np.tan(self.beta)) - self.L_y + self.h0

        self.x_circle = self.parent.length / 2
        self.y_circle_up = -self.r_in * np.cos(self.alpha) + (0.5 * self.parent.length - self.r_in * np.sin(self.alpha)) * np.tan(
            self.alpha) + self.h0
        self.y_circle_dw = (0.5 * self.parent.length - self.r_in * np.sin(self.beta)) * np.tan(self.beta) - self.r_in * np.cos(
            self.beta)

        self.up_x1, self.up_y1 = [self.parent.length / 2 - self.r_in * np.sin(self.alpha),
                                  (self.parent.length / 2 - self.r_in * np.sin(self.alpha)) * np.tan(self.alpha) + self.h0]
        self.dwn_x1, self.dwn_y1 = [self.parent.length / 2 - self.r_in * np.sin(self.beta),
                                    (self.parent.length / 2 - self.r_in * np.sin(self.beta)) * np.tan(self.beta)]
        self.up_x2, self.up_y2 = [self.parent.length / 2 + self.r_in * np.sin(self.alpha),
                                  (self.parent.length / 2 - self.r_in * np.sin(self.alpha)) * np.tan(self.alpha) + self.h0]
        self.dwn_x2, self.dwn_y2 = [self.parent.length / 2 + self.r_in * np.sin(self.beta),
                                    (self.parent.length / 2 - self.r_in * np.sin(self.beta)) * np.tan(self.beta)]
        self.hap = self.get_H(0.5)

    def f1(self, x: float) -> float:
        return np.tan(self.alpha) * x + self.h0

    def f2(self, x: float) -> float:
        return np.tan(self.beta) * x

    def f3(self, x: float) -> float:
        return (self.parent.length - x) * np.tan(self.alpha) + self.h0

    def f4(self, x: float) -> float:
        return 2 * self.L_y - np.tan(self.beta) * x

    def f5(self, x: float) -> float:
        return np.sqrt(self.r_in ** 2 - (self.parent.length / 2 - x) ** 2) + self.y_circle_up

    def f6(self, x: float) -> float:
        return np.sqrt(self.r_in ** 2 - (self.parent.length / 2 - x) ** 2) + self.y_circle_dw

    @property
    def H(self) -> float:
        return self.get_H(0)

    @property
    def H0(self) -> float:
        return self.h0

    @property
    def Hap(self) -> float:
        return self.hap

    @H0.setter
    def H0(self, val: (int, float)):
        self.h0 = val
        self.update_circles()
        self.update_elements()

    @property
    def Alpha(self) -> float:
        return self.alpha

    @Alpha.setter
    def Alpha(self, val: (int, float)):
        self.alpha = np.pi * val / 180
        self.update_circles()
        self.update_elements()

    @property
    def Beta(self) -> float:
        return self.beta

    @Beta.setter
    def Beta(self, val: (int, float)):
        self.beta = np.pi * val / 180
        self.update_circles()
        self.update_elements()

    @property
    def R_in(self) -> float:
        return self.r_in

    @R_in.setter
    def R_in(self, val: (int, float)):
        self.r_in = val
        self.update_circles()
        self.update_elements()

    def get_unmodified_h(self, x: (int, float)) -> float:
        x *= self.parent.length
        if 0 <= x < self.dwn_x1:
            h = self.f1(x) - self.f2(x)
        elif self.dwn_x1 <= x < self.parent.length / 2:
            h = self.f1(x) - self.f6(x)
        elif self.parent.length / 2 <= x < self.dwn_x2:
            h = self.f3(x) - self.f6(x)
        else:
            h = self.f3(x) - self.f4(x)
        return h

    def A_face(self):
        if1 = integrate.quad(self.f1, 0, self.parent.length/2)
        if2 = integrate.quad(self.f2, 0, self.dwn_x1)
        if3 = integrate.quad(self.f3, self.parent.length/2, self.parent.length)
        if4 = integrate.quad(self.f4, self.dwn_x2, self.parent.length)
        if6 = integrate.quad(self.f6, self.dwn_x1, self.dwn_x2)
        self.intf1 = if1[0]
        self.intf2 = if2[0]
        self.intf3 = if3[0]
        self.intf4 = if4[0]
        self.intf6 = if6[0]
        a = if1[0] + if3[0] - if2[0] - if4[0] - if6[0]

        return a

    def __str__(self):
        return f'KaarevaHarjaPalkki({round(self._B, 0)}X{round(self.H0, 0)}X{round(self.Alpha, 0)}X{round(self.Beta, 0)}X{round(self.R_in, 0)}X{round(self.Hap, 0)})'


class SmallElementSection:
    def __init__(self, A: (int, float), I: list_out):
        self.A = A
        self.I = I

    def __repr__(self):
        return f'SmallElementSection A:{self.A}, I:{self.I}'


if __name__ == '__main__':
    import frame2d.frame2d as f2d
    from time import time

    ts = TimberSection(30, 50)
    pp = PulpettiPalkki(200, 300, 500)
    hp = HarjaPalkki(20, 30, 100)
    mp = MahaPalkki(100, 200, 3)
    kp = KaarevaPalkki(100, 600, 20, 15, 15000)
    khp = KaarevaHarjaPalkki(100, 600, 20, 15, 15000)
    obj = type('', (), {})()
    obj.length = 10000
    khp.parent = obj
    kp.parent = obj
    khp.update_circles()
    kp.update_circles()

    h = []
    i_l = []
    start = time()
    for i in np.arange(1 + 0.001, step=0.001):
        # print(f'{i}   h {t.get_H(i)}')
        #print(pp.get_A(i))
        h.append(khp.f4(i))
        i_l.append(i)
        # print(f'H: {t.get_H(i)}')
        # print(f'A: {t.get_A(i)}')
        # print(f'I: {t.get_I(i)}')
        # print(f'Wel: {t.get_Wel(i)}')
    print(f'aikaa kulunut {round(time() - start, 20)} s')
    print(f'laskentakertoja {len(h)}')
    plt.plot(i_l, h)
    plt.show()



__all__ = ["FrameFEM", "Material", "Section", "BeamSection",
           "Support", "Load", "PointLoad", "LineLoad", "LoadCase",
           "FEMNode", "Element"]
#from .framefem import FrameFEM, BeamSection, PointLoad, LineLoad, Element, Section, Material
from .framefem import *
name = "framefem"




# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela

from metku.frame2d.frame2d import *
from metku.optimization.structopt import *

FIFTEEN_BAR_AREAS_in2 = [0.111, 0.141, 0.174, 0.220, 0.270, 0.287, 0.347,
                         0.440, 0.539, 0.954, 1.081, 1.174, 1.333, 1.488,
                         1.764, 2.142, 2.697, 2.800, 3.131, 3.565, 3.813,
                         4.805, 5.952, 6.572, 7.192, 8.525, 9.300, 10.850,
                         13.300, 14.290, 17.170, 19.180]

FIFTEEN_BAR_AREAS_mm2 = [71.613, 90.968, 112.258, 141.935, 174.193, 185.161,
                         223.871, 283.87, 347.741, 615.483, 697.418, 757.418,
                         859.998, 959.998, 1138.062, 1381.933, 1739.997,
                         1806.448, 2019.996, 2299.995, 2459.995, 3099.994,
                         3839.992, 4239.992, 4639.991, 5499.989, 5999.988,
                         6999.986, 8580.628, 9219.336, 11077.397, 12374.169]



class FifteenBarTruss(OptimizationProblem):
    # Problem parameters
    L = 3048  # mm
    F = 44537 # N # 1e5 lb
    E = 68950  # MPa # 10e7 psi
    rho = 2720e-9  # kg/mm3

    properties = {
        'L': str(L) + " mm",
        'F': f'{F / 1e3 :.2f} kN',
        'E': f'{E / 1e3 :.2f} MPa',
        'rho': f'{rho * 1e9 :.0f} kg/m3'
    }
    # Constraint limits

    sigma_max = 172.37  # MPa # 25000 psi
    delta_max = 50.8  # mm

    def __init__(self, prob_type='discrete'):
        super().__init__(name='FifteenBarTruss')
        self.prob_type = prob_type
        self.structure = self.create_structure()
        self.create_variables(profiles=FIFTEEN_BAR_AREAS_mm2)
        self.create_constraints()
        self.create_objective()

    def create_objective(self):

        def objective(X):
            self.substitute_variables(X)
            weight = 0
            for mem in self.structure.members.values():
                weight += self.rho * mem.A * mem.length

            return weight

        self.obj = objective

    def create_variables(self, profiles=[0, 1e6]):
        """
        Creates variables used in optimization

        Appends each variable to 'vars' 'list from where they can be accessed
        """
        self.vars = []
        # Member area variables
        for i, mem in enumerate(self.structure.members.values()):
            name = 'A' + str(i + 1)
            if self.prob_type == "discrete":
                var = DiscreteVariable(name,
                                       values=profiles,
                                       target={"property": "A",
                                               "objects": [mem]})

            elif self.prob_type == "continuous":
                var = Variable(name,
                               lb=profiles[0],
                               ub=profiles[-1],
                               target={"property": "A",
                                       "objects": [mem]})
            else:

                raise TypeError("Problem type must be either 'dicrete' "
                                "or 'continuous")

            self.add(var)

        # Node location variables
        # Nodes 2, 3, 6, 7 can move in x- and y -directions
        # x2 == x6 , x3 == x7
        # Nodes 4, 8 can only move in y -direction
        # Nodes 1, 5 are stationary
        x_nodes = [self.structure.f.nodes[i] for i in [1, 2, 5, 6]]
        y_nodes = [self.structure.f.nodes[i] for i in [1, 2, 3, 5, 6, 7]]

        x2 = Variable('x2',
                      lb=254,
                      ub=3556,
                      target={"property": "x",
                              "objects": [x_nodes[0], x_nodes[2]]})

        x3 = Variable('x3',
                      lb=5588,
                      ub=6604,
                      target={"property": "x",
                              "objects": [x_nodes[1], x_nodes[3]]})

        y2 = Variable('y2',
                      lb=2540,
                      ub=3556,
                      target={"property": "y",
                              "objects": [x_nodes[0]]})

        y3 = Variable('y3',
                      lb=2540,
                      ub=3556,
                      target={"property": "y",
                              "objects": [y_nodes[1]]})
        y4 = Variable('y4',
                      lb=1525,
                      ub=2286,
                      target={"property": "y",
                              "objects": [y_nodes[2]]})
        y6 = Variable('y6',
                      lb=-508,
                      ub=508,
                      target={"property": "y",
                              "objects": [y_nodes[3]]})
        y7 = Variable('y7',
                      lb=-508,
                      ub=508,
                      target={"property": "y",
                              "objects": [y_nodes[4]]})
        y8 = Variable('y8',
                      lb=508,
                      ub=1524,
                      target={"property": "y",
                              "objects": [y_nodes[5]]})

        for var in [x2, x3, y2, y3, y4, y6, y7, y8]:
            self.add(var)

    def create_structure(self):

        frame = Frame2D(num_elements=1)
        # Nodes
        n1 = [0, self.L]
        n2 = [self.L, self.L]
        n3 = [2 * self.L, self.L]
        n4 = [3 * self.L, self.L]
        n5 = [0, 0]
        n6 = [self.L, 0]
        n7 = [2 * self.L, 0]
        n8 = [3 * self.L, 0]
        # Create bars
        bar1 = SteelBeam([n1, n2])
        bar2 = SteelBeam([n2, n3])
        bar3 = SteelBeam([n3, n4])
        bar4 = SteelBeam([n5, n6])
        bar5 = SteelBeam([n6, n7])
        bar6 = SteelBeam([n7, n8])
        bar7 = SteelBeam([n6, n2])
        bar8 = SteelBeam([n7, n3])
        bar9 = SteelBeam([n8, n4])
        bar10 = SteelBeam([n1, n6])
        bar11 = SteelBeam([n5, n2])
        bar12 = SteelBeam([n2, n7])
        bar13 = SteelBeam([n6, n3])
        bar14 = SteelBeam([n3, n8])
        bar15 = SteelBeam([n7, n4])
        # Add bars
        frame.add(bar1)
        frame.add(bar2)
        frame.add(bar3)
        frame.add(bar4)
        frame.add(bar5)
        frame.add(bar6)
        frame.add(bar7)
        frame.add(bar8)
        frame.add(bar9)
        frame.add(bar10)
        frame.add(bar11)
        frame.add(bar12)
        frame.add(bar13)
        frame.add(bar14)
        frame.add(bar15)
        # Supports
        frame.add(XYHingedSupport(n1))
        frame.add(XYHingedSupport(n5))
        # Point loads
        frame.add(PointLoad(n8, [0, -self.F, 0]))
        # Change material properties
        for mem in frame.members.values():
            mem.fy = self.sigma_max
            mem.E = self.E  # MPa
            mem.A = FIFTEEN_BAR_AREAS_mm2[0]
            mem.Sj1 = 0
            mem.Sj2 = 0
        frame.generate()
        frame.calculate()
        return frame


    def constraint_generator(self, mem):

        def tension_fun(X):
            return mem.ned / mem.NRd - 1

        def compression_fun(X):
            return -mem.ned / mem.NRd - 1

        def buckling_fun(X):
            sigma_cr = 100 * mem.E * mem.A / (8 * mem.length ** 2)
            sigma = -mem.ned / mem.A
            return sigma / sigma_cr - 1


        return tension_fun, compression_fun, buckling_fun

    def create_constraints(self):

        # Initialize constraints as an empty list
        self.cons = []

        i = 0
        for j, var in enumerate(self.vars):
            for mem in var.target["objects"]:
                if isinstance(mem, FrameMember):
                    i += 1

                    compression_fun, tension_fun, buckling_fun = self.constraint_generator(mem)

                    comp_con = NonLinearConstraint(con_fun=compression_fun,
                                                   name="Compression " + str(
                                                       i),
                                                   problem=self)
                    comp_con.fea_required = True

                    tension_con = NonLinearConstraint(con_fun=tension_fun,
                                                      name="Tension " + str(i),
                                                      problem=self)
                    tension_con.fea_required = True

                    # buckl_con = NonLinearConstraint(con_fun=buckling_fun,
                    #                                name='Buckling ' + str(
                    #                                    i),
                    #                                problem=self)
                    # buckl_con.fea_required = True

                    self.add(comp_con)
                    self.add(tension_con)
                    # self.add(buckl_con)



if __name__ == '__main__':
    from metku.optimization.solvers import *
    problem = FifteenBarTruss(prob_type='continuous')
    #solver = GA(popsize=10)
    solver = SLP(move_limits=[0.05, 0.05], gamma=1e-2)
    x0 = [var.ub for var in problem.vars]
    solver.solve(problem, x0=x0, maxiter=200, verb=True, plot=True)
    problem(solver.X)
    problem.structure.plot()



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Thu Mar 15 18:12:04 2018

Rod element, for pin-jointed members with only axial forces

@author: kmela
"""

import numpy as np

try:
    from metku.framefem import Element
except:
    from framefem import Element
        


class Rod(Element):
    """ Rod element, carries only axial loads """

    def __init__(self, n1, n2, section, material):
        super().__init__(n1, n2, section, material)
                
        # Dimension of the element
        self.dim = len(n1.coord)
        self.L = self.transformation_matrix()

    def transformation_matrix(self,initial=True):
        """ From local to global coordinates """
        c = self.direction_cosines()
        #c = self.dcos
        L1 = np.hstack((c, np.zeros(len(c))))
        L2 = np.hstack((np.zeros(len(c)), c))
        L = np.vstack((L1, L2))
        return L

    def global_dofs(self):
        """ Get numbering of element global degrees of freedom """
        d = self.dim            
        return np.append(self.nodes[0].dofs[:d],self.nodes[1].dofs[:d])
        #return self.nodes[0].dofs[:d] + self.nodes[1].dofs[:d]


    def init_dofs(self):
        """ For truss members there are no rotational degrees of freedom
            so these can be set to be neglected (=1)
        """
        for n in self.nodes:
            n.dofs[0] = 0
            n.dofs[1] = 0
            
            if self.dim == 3:
                n.dofs[2] = 0
            
            """
            if self.dim == 2:
                n.dofs[2] = 1                
            else:
                n.dofs[3] =  1
                n.dofs[4] =  1
                n.dofs[5] =  1
            """
            
            

    def local_stiffness_matrix(self):
        """ Stiffness matrix in local coordinates """
        E = self.material.young
        A = self.section.A
        Le = self.length()
        #Le = self.len

        """ Stiffness matrix in local coordinates """
        return E * A / Le * np.array([[1, -1], [-1, 1]])

    def stiffness_matrix(self):
        """ Compute the stiffness matrix """
        k0 = self.local_stiffness_matrix()

        """ Transformation matrix """
        L = self.transformation_matrix()
        
        """ Transform stiffness matrix to global coordinates """
        ke = L.transpose().dot(k0.dot(L))

        return ke

    def local_geometric_stiffness_matrix(self,lcase=0):
        """ Geometric stiffness matrix in local coordinates
            From: Cook et. al 1989, Section 14.2
            
            TODO: Modify for 3D!
        """
        #P = self.axial_force[1]
        P = self.find[lcase]['fx'][0]
        Le = self.length()

        return P / Le * np.array([[0, 0, 0, 0], [0, 1, 0, -1], [0, 0, 0, 0], [0, -1, 0, 1]])

    def geometric_stiffness_matrix(self,lcase=0):
        """ Geometric stiffness matrix in global coordinates
            From: Cook et. al (1989), Section 14.2
        """
        """ Generate first the transformation matrix
            that takes into account two DOFs per node
            From: Cook et. al (1989), Section 7.5
            
            TODO: Modify for 3D!
        """
        c = self.direction_cosines()
        L = np.array([[c, 0, 0], [-c[1], c[0], 0, 0], [0, 0, c], [0, 0, -c[1], c[0]]])

        kG0 = self.local_geometric_stiffness_matrix(lcase)

        kG = L.transpose().dot(kG0.dot(L))

        return kG

    def equivalent_nodal_loads(self, q):
        """ Equivalent nodal loads for load in vector q
            Ignore bending moment
            
            TODO: Modify for 3D!
        """

        fglob = np.zeros(4)

        dc = self.direction_cosines()
        L = self.length()
        
        #dc = self.dcos
        #L = self.len

        fglob[[0, 2]] = 0.5 * dc[1] * q[0] * L
        fglob[[1, 3]] = 0.5 * dc[0] * q[1] * L

        return fglob


    def internal_forces(self, lcase=0, initial=True):
        """ Calculates internal forces (axial force)
        """
        q = self.local_displacements(lcase)
        E = self.material.young
        A = self.section.A
        L = self.length()
        #L = self.len
        self.fint[lcase]['fx'][0] = E * A / L * (q[1] - q[0])
        self.fint[lcase]['fx'][1] = self.fint[lcase]['fx'][0]


    def nodal_displacements(self,lcase=0, order=1):
        """ Get nodal displacements of an element
            Requires previously performed structural analysis such
            that nodal displacements are available.
        """

        d = self.dim
        return np.hstack((self.nodes[0].u[lcase][:d], self.nodes[1].u[lcase][:d]))


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 14 22:50:51 2018

Constants for EN 1992

@author: kmela
"""

from math import sqrt


# Partial safety factors
gammaC = 1.5    # Concrete
gammaS = 1.15   # Steel

# EN 1992-1-1 3.1.6
alpha_cc = 0.85 


""" Material data
fck .. lieriölujuus
fck_cube .. kuutiolujuus
fcm
Ecm
"""
concrete = {"C12/15": {"fck": 12.0, "fck_cube": 15.0, "fcm": 20, "Ecm": 27000.0},
            "C16/20": {"fck": 16.0, "fck_cube": 20.0, "fcm": 24, "Ecm": 29000.0},
            "C20/25": {"fck": 20.0, "fck_cube": 25.0, "fcm": 28, "Ecm": 30000.0},
            "C25/30": {"fck": 25.0, "fck_cube": 30.0, "fcm": 33, "Ecm": 31000.0},
            "C30/37": {"fck": 30.0, "fck_cube": 37.0, "fcm": 38, "Ecm": 33000.0},
            "C35/45": {"fck": 35.0, "fck_cube": 45.0, "fcm": 43, "Ecm": 34000.0},
            "C40/50": {"fck": 40.0, "fck_cube": 50.0, "fcm": 48, "Ecm": 35000.0},
            "C45/55": {"fck": 45.0, "fck_cube": 55.0, "fcm": 53, "Ecm": 36000.0},
            "C50/60": {"fck": 50.0, "fck_cube": 60.0, "fcm": 58, "Ecm": 37000.0},
            }

class Concrete:
    
    def __init__(self,material="C25/30"):
        """ Constructor for Concrete material class """

        self.name = material
        self.fck = concrete[material]["fck"]
        self.fck_cube = concrete[material]["fck_cube"]
        
        self.fcm = self.fck + 8.0
        self.Ecm = 22000*(self.fcm/10)**0.3
        
        # density [kn/m3], from EN 1991-1-1, Table A.1
        self.density = 24
        self.RH = 50
        
        #self.fcm = concrete[material]["fcm"]
        #self.Ecm = concrete[material]["Ecm"]
        
    def fcd(self,a_cc=alpha_cc,gC=gammaC):
        """ Design value for compressive strength """
        return a_cc*self.fck/gC
    
    def nu_cracked(self):
        """ Reduction factor for concrete cracked due to shear """
        return 0.6*(1-self.fck/250)
    
    def creep_coefficient(self,h0,t,t0=None):
        """ According to EN 1992-1-1, Annex B """
        
        RH = self.RH
        
        """ Eq. (B.8c) """
        alpha = [0,0,0]
        alpha[0] = (35/self.fcm)**0.7
        alpha[1] = (35/self.fcm)**0.2
        alpha[2] = (35/self.fcm)**0.5
        
        """ Eq. (B.3) """
        if self.fcm <= 35:
            phiRH = 1 + (1-RH/100)/0.1/h0**(1/3)
            betaH =  min(1.5*(1+(0.012*RH)**18)*h0+250,1500)
        else:
            phiRH = (1 + (1-RH/100)/0.1/h0**(1/3)*alpha[0])*alpha[1]
            betaH =  min(1.5*(1+(0.012*RH)**18)*h0+250*alpha[2],1500*alpha[2])
        
        """ Eq. (B.4) """
        beta_cm = 16.8/sqrt(self.fcm)
        
        """ Eq. (B.5) """
        beta_t0 = 1/(0.1+t0**0.2)
        
        """ Eq. (B.2) """
        phi0 = phiRH*beta_cm*beta_t0
                
        
        """ Eq. (B.7) """
        beta_c = ((t-t0)/(betaH + t-t0))**0.3
                
        phi = phi0 * beta_c
        
        return phi


#  This source code is licensed under the MIT license. See LICENSE in the repository root directory.
#  Copyright (c) 2024. Metku team.
#  All rights reserved.



import math
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as lines
import os

from metku.sections.steel.ISection import HEA, HEB, IPE
from metku.structures.steel.plates import RectPlate, RectPlateWithHoles
from metku.eurocodes.en1993.constants import gammaM0, gammaM1, gammaM2
from metku.eurocodes.en1993.en1993_1_1 import buckling_reduction_factor, buckling_curve
from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import Bolt, BoltRow, BoltRowGroup, stiffness_ratio, full_strength_weld_throat
from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import nut_size, washer_size
import metku.eurocodes.en1993.en1993_1_8.component_method as cm

from metku.cost.workshop import Workshop
from metku.cost.cost_data import BASIC_STEEL_PRICE, steel_grade_add_on, thickness_add_on
from metku.cost.cost_data import bolt_unit_cost_full_thread as bolt_costs_ft
from metku.cost.cost_data import bolt_unit_cost_part_thread as bolt_costs_pt
from metku.cost.cost_data import nut_unit_cost, washer_unit_cost


class FinPlateJoint:
    """ Class for fin plate joints """
    def __init__(self, beam, plate, bolt=Bolt(20,8.8), gh=10,workshop=None):
        """
            beam .. beam profile (ISection)
            plate .. RectPlateWithHoles object
            bolt .. Bolt class object for bolts
            gh .. gap between beam end and support
            
            -- TWO of these three should be given --
            hp .. height of fin plate
            e1 .. vertical edge distance
            p1 .. vertical distance between bolt centers
            
            
            bp .. width of fin plate            
            z .. horizontal distance from support to bolt group center
            e2 .. horizontal edge distance
            p2 .. horizontal distance between bolt centers
            
        """
        self.beam = beam
        self.plate = plate
        # Vertical position of the center of the plate relative to
        # centerline of the beam.
        self.plate_pos_y = 0.0
        self.bolt = bolt
        self.plate.d0 = bolt.d0
        self.gh = gh
        self.workshop = workshop
        
        self.weld_size = 5.0 # throat thickness of a fillet weld, mm
        
        self._cost = {'plate':0.0,                                         
                     'welding':0.0,
                     'beam holes':0.0,
                     'bolts':0.0}
    
        self.vertical_symmetry = True
    
    @property
    def e1(self):
        """ Vertical edge distance to fin plate edge """
        if self.vertical_symmetry:
            # Here, we require that the vertical edge distances are the same
            # from the top and from the bottom of the plate
            self.plate.x0[1] = 0.5*self.hp - 0.5*(self.plate.n1-1)*self.p1
        return self.plate.eTop()
    
    @property
    def e2(self):
        """ Horizontal edge distance to fin plate edge """
        return self.plate.eRight()
    
    @property
    def e1b(self):
        """ Vertical edge distance to beam end edge """
        return 0.5*self.beam.h-(self.plate_pos_y+0.5*self.hp-self.e1)
    
    @property
    def e2b(self):
        """ Horizontal edge distance to beam end edge """
        return self.plate.eLeft()-self.gh
        
    @property
    def z(self):
        """ Horizontal distance from support to bolt group center """
        return self.plate.eLeft() + 0.5*(self.plate.n2-1)*self.plate.px
    
    @property
    def bp(self):
        return self.plate.b
    
    @bp.setter
    def bp(self,value):
        self.plate.b = value
    
    @property
    def hp(self):
        return self.plate.h

    @hp.setter 
    def hp(self,value):
        self.plate.h = value
    
    @property
    def tp(self):
        return self.plate.t
    
    @tp.setter
    def tp(self,value):
        self.plate.t = value
        
    @property
    def fyp(self):
        """ Yield strength of fin plate """
        return self.plate.fy
    
    @property
    def fup(self):
        """ Ultimate strength of fin plate """
        return self.plate.fu
    
    @property
    def d0(self):
        """ Bolt hole diameter """
        return self.bolt.d0
        
    @property
    def p1(self):
        """ Vertical distance between bolt hole centers """
        return self.plate.py

    @p1.setter
    def p1(self,value):
        """ Vertical distance between bolt hole centers """
        self.plate.py = value

    @property
    def p2(self):
        """ Horizontal distance between bolt hole centers """
        return self.plate.px
    
    @p2.setter
    def p2(self,value):
        """ Vertical distance between bolt hole centers """
        self.plate.px = value
    
    @property
    def I(self):
        """ Moment of inertia of bolt row group with two vertical rows """
        return 0.5*self.plate.n1*self.p2**2 + 1/6*self.plate.n1*(self.plate.n1**2-1)*self.p1**2
    
    def bolt_shear(self,verb=False):
        """ Shear resistance of bolts """
        n = self.plate.n
        z = self.z
        p1 = self.p1
        
        FvRd = self.bolt.shear_resistance()
        
        if self.plate.n2 == 1:
            C = n/np.sqrt(1+(6*z/(n+1)/p1)**2)
        else:
            p2 = self.p2
            Ibolts = self.I
            n1 = self.plate.n1
            C = 1/np.sqrt((0.5*z*p2/Ibolts+1/n)**2 + (0.5*z*p1/Ibolts*(n1-1))**2)
        
        VRd1 = C*FvRd
        
        if verb:
            print(f'VRd1 = {VRd1*1e-3:4.2f} kN')
        
        return VRd1
    
    def fin_plate_bearing(self,verb=False):
        """ Bearing resistance of fin plate """
        n1 = self.plate.n1
        if self.plate.n2 == 1:
            p2 = 1e5
            a = 0
            b = 6*self.z/n1/(n1+1)/self.p1  
        else:
            p2 = self.p2
            a = 0.5*self.z*p2/self.I
            b = 0.5*self.z*self.p1/self.I*(n1-1)
        
        ever = [self.e1,self.e2]        
        pver = [self.p1,p2]
        # NOTE! This bearing resistance distinguishes between edge and other bolts.
        # ECCS guide takes the minimum edge distance term using e and p regardless of
        # bolt position.
        FbverRd = self.bolt.bearing_resistance(self.fup, self.tp, ever, pver)
        
        
        ehor = [self.e2,self.e1]
        phor = [p2,self.p1]
        FbhorRd = self.bolt.bearing_resistance(self.fup, self.tp, ehor, phor)
        
                
        n = self.plate.n
        VRd2 = n/np.sqrt(((1+n*a)/FbverRd)**2+(n*b/FbhorRd)**2)
        
        if verb:
            print('Fin plate bearing resistance:')
            print(f'Fb,ver,Rd = {FbverRd*1e-3:4.2f} kN')
            print(f'Fb,hor,Rd = {FbhorRd*1e-3:4.2f} kN')
            print(f'     alpha = {a:4.2f}')
            print(f'     beta  = {b:4.2f}')
            print(f'     VRd2 = {VRd2*1e-3:4.2f} kN')
        
        return VRd2
    
    def fin_plate_shear_gross(self,verb=False):
        """ Shear resistance of fin plate """
        
        VRd3 = self.hp*self.tp*self.fyp/1.27/np.sqrt(3)/gammaM0

        if verb:
            print(f'Fin plate shear resistance, gross section: VRd3 = {VRd3*1e-3:4.2f} kN')
        
        return VRd3
    
    def fin_plate_shear_net(self,verb=False):
        """ Shear resistance of fin plate, net section """
        
        Avnet = self.tp*(self.hp-self.plate.n1*self.d0)
        VRd4 = Avnet*self.fup/np.sqrt(3)/gammaM2

        if verb:
            print(f'Fin plate shear resistance, net section: VRd4 = {VRd4*1e-3:4.2f} kN')
        
        return VRd4

    def fin_plate_block_tearing(self,verb=False):
        """ Shear resistance of fin plate, block tearing """
        
        Ant = self.tp*(self.e2+self.p2-(2*self.plate.n2-1)*0.5*self.d0)
        Anv = self.tp*(self.hp-self.e1-(self.plate.n1-0.5)*self.d0)
        VRd5 = 0.5*self.fup*Ant/gammaM2 + self.fyp*Anv/np.sqrt(3)/gammaM0

        if verb:
            print(f'Fin plate shear resistance, block tearing: VRd5 = {VRd5*1e-3:4.2f} kN')
        
        return VRd5
    
    def fin_plate_bending(self,verb=False):
        """ Fin plate in bending resistance """
        
        Wel = self.tp*self.hp**2/6
        VRd6 = self.fyp*Wel/self.z/gammaM0
    
        if verb:
            print(f'Fin plate in bending: VRd6 = {VRd6*1e-3:4.2f} kN')
            if self.hp > 2.73*self.z:
                print(f'Condition hp = {self.hp:4.2f} mm >= 2.73*z = {2.73*self.z:4.2f} mm is satisfied, plate bending not relevant.')
            else:
                print('Condition hp >= 2.73*z is not satisfied, plate bending is relevant.')            
                print(f'hp = {self.hp:4.2f} mm, 2.73*z = {2.73*self.z:4.2f} mm')
        
        return VRd6
    
    def fin_plate_ltb(self,verb=False):
        """ Lateral-torsional buckling of fin plate """
        
        VpRd = self.fyp*self.tp*self.hp**2/6/self.z
        
        if self.z > self.tp/0.15:      
            if verb:
                print('Fin plate LTB need not be checked.')
            C = 1.0
        else:
            if verb:
                print('Fin plate LTB needs to be checked.')
            l1 = np.pi*np.sqrt(self.plate.E/self.fyp)
            lLT = 2.8/l1*np.sqrt(self.z*self.hp/1.5/(self.tp**2))
            
            # !TODO: Check how chi_LT should really be determined!
            chiLT = buckling_reduction_factor(lLT,buckling_curve['d'])
            C =  min(chiLT/0.6/gammaM1,1/gammaM0)
            
        VbRd = C*VpRd
        
        if verb:
            print(f'Fin plate LTB resistance: VRd7 = {VbRd*1e-3:4.2f} kN')
            print(f'                          VpRd = {VpRd*1e-3:4.2f} kN')
            print(f'                            l1 = {l1:4.2f}')
            print(f'                           lLT = {lLT:4.2f}')
            print(f'                             C = {C:4.2f}')
        
        return VbRd
    
    def beam_web_bearing(self,verb=False):
        """ Beam web bearing resistance """
        
        n1 = self.plate.n1
        if self.plate.n2 == 1:
            p2 = 1e5
            a = 0
            b = 6*self.z/n1/(n1+1)/self.p1  
        else:
            p2 = self.p2
            a = 0.5*self.z*p2/self.I
            b = 0.5*self.z*self.p1/self.I*(n1-1)
        
        ever = [self.e1,self.e2b]        
        pver = [self.p1,p2]
        # NOTE! This bearing resistance distinguishes between edge and other bolts.
        # ECCS guide takes the minimum edge distance term using e and p regardless of
        # bolt position.
        FbverRd = self.bolt.bearing_resistance(self.beam.fu, self.beam.tw, ever, pver)
        
        
        ehor = [self.e2b,self.e1]
        phor = [p2,self.p1]
        FbhorRd = self.bolt.bearing_resistance(self.beam.fu, self.beam.tw, ehor, phor)
        
                
        n = self.plate.n
        VRd8 = n/np.sqrt(((1+n*a)/FbverRd)**2+(n*b/FbhorRd)**2)
        
        if verb:
            print('Beam web bearing resistance:')
            print(f'Fb,ver,Rd = {FbverRd*1e-3:4.2f} kN')
            print(f'Fb,hor,Rd = {FbhorRd*1e-3:4.2f} kN')
            print(f'     alpha = {a:4.2f}')
            print(f'     beta  = {b:4.2f}')
            print(f'     VRd8 = {VRd8*1e-3:4.2f} kN')
        
        return VRd8
    
    def beam_web_shear_gross(self,verb=False):
        """ Shear resistance of beam web """
        
        VRd9 = self.beam.VzRd

        if verb:
            print(f'Beam shear resistance, gross section: VRd9 = {VRd9*1e-3:4.2f} kN')
        
        return VRd9
    
    def beam_web_shear_net(self,verb=False):
        """ Shear resistance of beam web, net section """
        
        Abvnet = self.beam.Ashear[1] - self.plate.n1*self.d0*self.beam.tw
        VRd10 = Abvnet*self.beam.fu/np.sqrt(3)/gammaM2

        if verb:
            print(f'Beam web shear resistance, net section: VRd10 = {VRd10*1e-3:4.2f} kN')
        
        return VRd10
    
    def beam_web_block_tearing(self,verb=False):
        """ Shear resistance of beam web, block tearing """
        
        Ant = self.beam.tw*(self.e2b+self.p2-(2*self.plate.n2-1)*0.5*self.d0)
        Anv = self.beam.tw*(self.e1b+(self.plate.n1-1)*self.p1-(self.plate.n1-0.5)*self.d0)
        VRd11 = 0.5*self.beam.fu*Ant/gammaM2 + self.beam.fy*Anv/np.sqrt(3)/gammaM0

        if verb:
            print(f'Beam shear resistance, block tearing: VRd11 = {VRd11*1e-3:4.2f} kN')
        
        return VRd11
    
    def draw(self,axes=None):
        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes
        
        h = self.beam.h
        tf = self.beam.tf
        bp = self.bp
        hp = self.hp
        gh = self.gh
        dx = 100
        dy = 20
        
        ax.vlines([0],-0.5*h-dy,0.5*h+dy,'k')
        
        ax.hlines([0.5*h,0.5*h-tf,-0.5*h,-0.5*h+tf],gh,gh+bp+dx,'k')
        ax.vlines([gh],0.5*h-tf,0.5*h,'k')
        ax.vlines([gh],-0.5*h,-0.5*h+tf,'k')
        
        # Origin of the coordinate system is at the centerline of the beam
        # and at the face of the support.
        x0 = (0,self.plate_pos_y-0.5*hp)
        self.plate.draw(x0,ax)
        
        ax.vlines([gh],0.5*hp,0.5*h-tf,'k')
        ax.vlines([gh],-0.5*h+tf,-0.5*hp,'k')
    
        plt.rcParams['hatch.color'] = 'k'
        hatch_rect = patches.Rectangle((-30,-0.5*h-dy), 30, h+2*dy, linewidth=0,hatch='/',edgecolor='k',facecolor='white')
        ax.add_patch(hatch_rect)
    
    def check_bolt_position(self,verb=False):
        """ Checks the bolt positioning rules of EN 1993-1-8 """
        res = True
        
        if self.e1 < 1.2*self.d0:
            if verb:
                print(f'Condition e1 >= 1.2*d0 is not satisfied: e1 = {self.e1:4.1g}, 1.2*d0 = {1.2*self.d0:4.1g}')
            res = False
        elif self.e2 < 1.2*self.d0:
            res = False
        elif self.e2b < 1.2*self.d0:
            res = False
        elif self.p1 < 2.2*self.d0:
            res = False
        elif self.p2 > 0 and self.p2 < 2.4*self.d0:
            res = False
        
        if res and verb:
            print('All bolt positioning conditions are OK!')
            
        return res
    
    def cost(self,workshop=None,material_cost=BASIC_STEEL_PRICE,verb=False):
        """ Calculate cost of the joint
            units: euro
            
            Total costs consist of:
                plate: including material, cutting, hole forming, painting
                welding: fin plate is welded to the column by two-sided fillet weld.
                beam holes: holes are drilled to the beam web
                bolts: material costs of bolts
                
            !TODO:
                Check beam hole forming, is it cutting, drilling or what
                Check welding, is it assembly welding or part welding.
        """
        
        for key, value in self._cost.items():
            self._cost[key] = 0.0
        
        if workshop is not None:
            self.workshop = workshop        
        
        plate_cost = self.workshop.steel_price["plates"]
        
        self._cost['plate'] = self.plate.cost(self.workshop,plate_cost)
        
        # Bolt costs: bolt assemply includes
        # 1) Bolt
        # 2) Nut
        # 3) One washer
        if self.bolt.bolt_type == "Full thread":
            self._cost['bolts'] += self.plate.n * bolt_costs_ft[int(self.bolt.length)][int(self.bolt.d)]
        else:
            self._cost['bolts'] += self.plate.n * bolt_costs_pt[int(self.bolt.length)][int(self.bolt.d)]
            
        self._cost['bolts'] += self.plate.n * nut_unit_cost[int(self.bolt.d)]
        self._cost['bolts'] += self.plate.n * washer_unit_cost[int(self.bolt.d)]
        
        hole_cut_length = self.plate.n * self.d0*math.pi
        
        self._cost['beam holes'] = self.workshop.cost_centres['cutting'].cost(self.beam.tw,hole_cut_length)

        """ Welding cost """
        plate_weld_length = 2*self.hp
        #print(flange_weld_length)
        self._cost['welding'] += self.workshop.cost_centres['assembly_welding'].cost(self.weld_size,plate_weld_length)
        
            
        total_cost = 0.0
        
        for c in self._cost.values():            
            total_cost += c
        
        self._total_cost = total_cost
        
        #self.cost_distribution()
        
        if verb:
            self.cost_distribution()
        
        return total_cost
    
    def cost_distribution(self,pie=False):
        """ Print cost distribution """
        
        if pie:
            pass
        else:
            print(""" Cost distribution of the connection """)
            p_tot = 0.0
            for key, value in self._cost.items():
                p =  value/self._total_cost*100
                p_tot += p
                print("   {0:10s}: {1:5.2f} € [{2:5.2f} %]".format(key,value,p))
            
            print("-----------------------")
            print("   {0:10s}: {1:5.2f} € [{2:5.2f} %]".format("Total",self._total_cost,p_tot))
    

def example():
    
    beam = IPE(300,fy=235)
    hp = 230
    e1 = 45
    n1 = 3
    n2 = 1
    b = Bolt(20,8.8)
    
    plate = RectPlateWithHoles(width=110, depth=230, thickness=10.0, d0=22, x0=[60,45],py=70,n2=1,material="S235")
    
    conn = FinPlateJoint(beam, plate, bolt=b)
    
    verb = True
    
    conn.check_bolt_position(verb)
    
    conn.bolt_shear(verb)
    conn.fin_plate_bearing(verb)
    conn.fin_plate_shear_gross(verb)
    conn.fin_plate_shear_net(verb)
    conn.fin_plate_block_tearing(verb)
    conn.fin_plate_bending(verb)
    conn.fin_plate_ltb(verb)
    conn.beam_web_bearing(verb)
    conn.beam_web_shear_gross(verb)
    conn.beam_web_shear_net(verb)
    conn.beam_web_block_tearing(verb)

    conn.workshop = Workshop()
    conn.cost(verb=verb)
    conn.draw()

    return conn


class FinPlateConnection:
    def __init__(self, plate_thickness, plate_width, plate_length, bolt_diameter, number_of_bolts):
        self.plate_thickness = plate_thickness
        self.plate_width = plate_width
        self.plate_length = plate_length
        self.bolt_diameter = bolt_diameter
        self.number_of_bolts = number_of_bolts

    def calculate_connection_details(self):
        # Implement the calculations to determine the connection details

        
        # just plot the plate dimensions
        pass



    def plot_connection_diagram(self):
        fig, ax = plt.subplots()
        ax.set_aspect('equal', adjustable='box')

        # Plot the fin plate
        ax.plot([0, self.plate_length], [0, 0], 'k-', linewidth=2)  # Bottom edge
        ax.plot([0, 0], [0, self.plate_width], 'k-', linewidth=2)   # Left edge
        ax.plot([self.plate_length, self.plate_length], [0, self.plate_width], 'k-', linewidth=2)  # Right edge
        ax.plot([0, self.plate_length], [self.plate_width, self.plate_width], 'k-', linewidth=2)  # Top edge

        # Plot the bolts (for simplicity, let's assume they're evenly spaced)
        bolt_spacing = self.plate_length / (self.number_of_bolts - 1)
        bolt_positions = [i * bolt_spacing for i in range(self.number_of_bolts)]
        for pos in bolt_positions:
            ax.plot([pos], [self.plate_width / 2], 'ro', markersize=8)  # Bolt position

        ax.set_xlabel('Length (mm)')
        ax.set_ylabel('Width (mm)')
        ax.set_title('Fin Plate Connection Diagram')

        plt.grid(True)
        plt.show()



# Example usage:
if __name__ == "__main__":
    plate_thickness = 10  # in mm
    plate_width = 150     # in mm
    plate_length = 300    # in mm
    bolt_diameter = 20    # in mm
    number_of_bolts = 3

    #connection = FinPlateConnection(plate_thickness, plate_width, plate_length, bolt_diameter, number_of_bolts)
    #connection.plot_connection_diagram()

    conn = example()    



















# Design Checks to be done in Fin plate connections

# Shear capacity of the bolt


#Tension capacity of the bolt


#Bearing capacity of the bolt


#Block shear capacity of the bolt


#Combined shear and tension capacity


#Weld check


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-

from metku.optimization.structopt import OptimizationProblem
from metku.optimization.variables import Variable, DiscreteVariable, BinaryVariable
from metku.optimization.constraints import NonLinearConstraint, LinearConstraint

from metku.frame2d.frame2d import *


THREE_BAR_AREAS_mm2 = [71.613, 90.968, 112.258, 141.935, 174.193, 185.161,
                         223.871,
                         283.87, 347.741, 615.483, 697.418, 757.418, 859.998,
                         959.998,
                         1138.062, 1381.933, 1739.997, 1806.448, 2019.996,
                         2299.995,
                         2459.995, 3099.994, 3839.992, 4239.992, 4639.991,
                         5499.989,
                         5999.988, 6999.986, 8580.628, 9219.336, 11077.397,
                         12374.169]


class ThreeBarTruss(OptimizationProblem):

    # Problem parameters
    L = 3048  # mm
    F = 444890  # N # 1e5 lb
    E = 68950  # MPa # 10e7 psi
    rho = 2768e-9  # kg/mm3

    properties = {
        'L': str(L) + " mm",
        'F': f'{F / 1e3 :.2f} kN',
        'E': f'{E / 1e3 :.2f} MPa',
        'rho': f'{rho * 1e9 :.0f} kg/m3'
    }

    # Constraint limits
    sigma_max = 172.37  # MPa # 25000 psi
    delta_max = 50.8  # mm

    def __init__(self, prob_type="discrete"):

        super().__init__(name='ThreeBarTruss')

        self.binary_vars = None
        self.prob_type = prob_type.lower()
        self.structure = self.create_structure()
        self.create_constraints()
        self.create_objective()

    def create_objective(self):

        def objective(X):
            self.substitute_variables(X)
            weight = 0
            for mem in self.structure.members.values():
                weight += self.rho * mem.A * mem.length  # mem.weight

            return weight

        self.obj = objective

    def create_structure(self):

        frame = Frame2D(num_elements=1)
        # Nodes
        n1 = [0, 0]
        n2 = [self.L, 0]
        n3 = [2 * self.L, 0]
        n4 = [self.L, -self.L]
        # Create bars
        bar1 = SteelBeam([n1, n4])
        bar2 = SteelBeam([n2, n4])
        bar3 = SteelBeam([n3, n4])
        # Add bars
        frame.add(bar1)
        frame.add(bar2)
        frame.add(bar3)
        # Supports
        frame.add(XYHingedSupport(n1))
        frame.add(XYHingedSupport(n2))
        frame.add(XYHingedSupport(n3))
        # Point loads
        frame.add(PointLoad(n4, [-self.F, -self.F, 0]))
        # Change material properties
        for mem in frame.members.values():
            mem.fy = self.sigma_max
            mem.E = self.E  # MPa
            mem.A = THREE_BAR_AREAS_mm2[0]
            mem.Sj1 = 0
            mem.Sj2 = 0
        frame.generate()
        frame.calculate()

        if self.prob_type == "discrete":
            # Variables
            A1 = DiscreteVariable('A1', profiles=THREE_BAR_AREAS_mm2,
                                  target={"property": "A",
                                          "objects": [bar1, bar3]})
            A2 = DiscreteVariable('A2', profiles=THREE_BAR_AREAS_mm2,
                                  target={"property": "A", "objects": [bar2]})
            vars = [A1, A2]

        elif self.prob_type == "continuous":
            A1 = Variable('A1', lb=THREE_BAR_AREAS_mm2[0],
                          ub=THREE_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar1, bar3]})
            A2 = Variable('A2', lb=THREE_BAR_AREAS_mm2[0],
                          ub=THREE_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar2]})

            vars = [A1, A2]


        elif self.prob_type == 'binary':
            cont_A1 = Variable('A1', lb=THREE_BAR_AREAS_mm2[0],
                          ub=THREE_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar1, bar3]})
            cont_A2 = Variable('A2', lb=THREE_BAR_AREAS_mm2[0],
                          ub=THREE_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar2]})


            binary_A1 = []
            for i in range(len(THREE_BAR_AREAS_mm2)):
                binary = BinaryVariable("Bin A1"+str(i))
                binary.target_val = THREE_BAR_AREAS_mm2[i]
                binary_A1.append(binary)

            binary_A2 = []
            for i in range(len(THREE_BAR_AREAS_mm2)):
                binary = BinaryVariable("Bin A2"+str(i))
                binary.target_val = THREE_BAR_AREAS_mm2[i]
                binary_A2.append(binary)

            vars = [cont_A1, cont_A2]
            vars.extend(binary_A1)
            vars.extend(binary_A2)
            self.binary_vars = [binary_A1, binary_A2]

        self.vars = vars

        return frame


    def binary_constraints(self):

        # All available areas
        A = np.zeros((len(self.binary_vars), (len(THREE_BAR_AREAS_mm2))))

        for i, binary_A in enumerate(self.binary_vars):
            # Binary constraint
            # sum(bin_vars) == 1
            binary_idx = [self.vars.index(bvar) for bvar in binary_A]
            a = np.zeros(len(self.vars))
            a[binary_idx] = 1
            bin_con = LinearConstraint(a=a, b=1, con_type="=",
                                       name="Binary Constraint "
                                            + str(i+1))
            self.cons.append(bin_con)

            # Binary Area constraint
            # Ai == sum(A[i][bin_idx])
            a[binary_idx] = [bvar.target_val for bvar in binary_A]
            b = self.vars[i]
            bin_A_con = LinearConstraint(a=a, b=b, con_type="=",
                                         name="Binary Area Constraint "
                                         + str(i + 1))
            self.cons.append(bin_A_con)


    def constraint_generator(self, mem):

        def tension_fun(X):
            return mem.ned / mem.NRd - 1

        def compression_fun(X):
            return -mem.ned / mem.NRd - 1

        def buckling_fun(X):
            sigma_cr = 100 * mem.E * mem.A / (8 * mem.length ** 2)
            sigma = -mem.ned / mem.A
            return sigma / sigma_cr - 1

        return tension_fun, compression_fun, buckling_fun

    def create_constraints(self):

        # Initialize constraints as an empty list
        self.cons = []

        i = 0
        for j, var in enumerate(self.vars):
            if not isinstance(var, BinaryVariable):
                for mem in var.target["objects"]:
                    if isinstance(mem, FrameMember):
                        i += 1

                        compression_fun, tension_fun, buckling_fun = self.constraint_generator(mem)

                        comp_con = NonLinearConstraint(con_fun=compression_fun,
                                                       name="Compression " + str(i),
                                                       problem=self)
                        comp_con.fea_required = True

                        tension_con = NonLinearConstraint(con_fun=tension_fun,
                                                          name="Tension " + str(i),
                                                          problem=self)
                        tension_con.fea_required = True

                        buckl_con = NonLinearConstraint(con_fun=buckling_fun,
                                                       name='Buckling ' + str(i),
                                                       problem=self)
                        buckl_con.fea_required = True

                        self.cons.append(comp_con)
                        self.cons.append(tension_con)
                        self.cons.append(buckl_con)

        if self.binary_vars:
            self.binary_constraints()

if __name__ == '__main__':
    from metku.optimization.solvers import *
    import matplotlib.pyplot as plt
    problem = ThreeBarTruss(prob_type='continuous')
    # X = [809, 7591, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    # problem(X)
    problem.structure.plot()



    """
    solver = SLP(move_limits=[0.95, 1.2])
    solver.problem = problem
    solver.random_feasible_point()
    x0 = [THREE_BAR_AREAS_mm2[-1], THREE_BAR_AREAS_mm2[-1]]
    solver.solve(problem,
                 maxiter=15,
                 maxtime=5,
                 log=True)
    problem(solver.X)
    xvals = np.asarray(solver.xvals)
    A1 = np.array(x0[0])
    A2 = np.array(x0[1])
    A1 = np.hstack((A1, xvals[:, 0]))
    A2 = np.hstack((A2, xvals[:, 1]))
    X, Y = np.meshgrid(A1, A2)
    Z1 = problem.structure.members[0].length * X * problem.rho
    Z2 = problem.structure.members[1].length * Y * problem.rho
    Z = 2*Z1 + Z2

    plt.plot(A1, A2)
    plt.scatter(A1, A2, c='k', marker='x')
    CS = plt.contour(X, Y, Z, levels=10, cmap='Greens_r')
    plt.clabel(CS, inline=True)

    plt.show()
    """